#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of Pymetrick.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.

"""Modulo para gestionar mail"""

try:
    from pymetrick.version import __author__, __version__, __license__, __copyright__
except ImportError:
    raise ImportError(
        'Failed to find (autogenerated) version.py '
    )

__date__ = '2014-05-21'
__credits__ = ''
__text__ = 'Gestion de funciones para Mail'
__file__ = 'mail.py'

#--- CHANGES ------------------------------------------------------------------
# 2015-05-04 v0.01 PL: - First version
# 2017-10-02 v0.43 PL: - Migrar a PYTHON 3.6

import sys
# sin bytecode
sys.dont_write_bytecode = True
import os
from os.path import basename, isfile
import re

import logging
import logging.handlers
import gc
import smtplib
import imaplib
try:
    import ssl
    from smtplib import SMTP, SMTP_SSL        # use this for standard SMTP protocol   (port 25, no encryption or port 465, uses SSL)
    from imaplib import IMAP4, IMAP4_SSL, Time2Internaldate
    IS_SSL = True
except:
    from smtplib import SMTP                  # use this for standard SMTP protocol   (port 25, no encryption or port 465, uses SSL)
    from imaplib import IMAP4,Time2Internaldate
    IS_SSL = False

import socket

# For guessing MIME type based on file name extension
import mimetypes

from optparse import OptionParser

import email
import mailbox
from email import encoders
from email.message import Message
from email.mime.base import MIMEBase
from email.mime.text import MIMEText
from email.mime.audio import MIMEAudio
from email.mime.image import MIMEImage
from email.mime.application import MIMEApplication
from email.mime.multipart import MIMEMultipart
from email.utils import COMMASPACE, formatdate

#from datetime import datetime, timedelta, tzinfo

from pymetrick.helpers import *
import time

# PY3
long = int

# Send an HTML email with an embedded image and a plain text message for
# email clients that don't want to display the HTML.

'''
DEBUG - debug message
INFO - info message
WARNING - warn message
ERROR - error message
CRITICAL - critical message
'''
LOG_LEVEL_VALUE = {'DEBUG':10,'INFO':20,'WARNING':30,'ERROR':40,'CRITICAL':50}
if str(os.environ.get('PYMETRICK_LOG_LEVEL',None)).upper() in LOG_LEVEL_VALUE:
    LOG_LEVEL = LOG_LEVEL_VALUE[str(os.environ.get('PYMETRICK_LOG_LEVEL')).upper()]
else:
    LOG_LEVEL = 30
LOG_FILENAME = '-'.join([os.path.abspath(__file__).split(os.sep)[len(os.path.abspath(__file__).split(os.sep))-1],])[:-3]
LOG = logging.getLogger(LOG_FILENAME)

if 'LD_LIBRARY_PATH' in list(os.environ.keys()):
    # CGI environment
    sys.stdout = sys.stderr
    logging.basicConfig(stream = sys.stderr, level=LOG_LEVEL, format='%(filename)s[line:%(lineno)d] - %(levelname)s - %(message)s')
else:
    # not CGI environment
    logging.basicConfig(stream=sys.stderr)
    hdlr = logging.handlers.RotatingFileHandler(filename=LOG_FILENAME+'.log',mode='a', encoding='utf-8', maxBytes=1048576, backupCount=3)
    formatter = logging.Formatter('%(asctime)s - %(filename)s[line:%(lineno)d] - %(levelname)s - %(message)s')
    hdlr.setFormatter(formatter)
    LOG.addHandler(hdlr)
    LOG.setLevel(LOG_LEVEL)


SMTP_DEFAULT_PORT = 25
SMTP_SSL_PORT = 465
SMTP_TLS_PORT = 587

IMAP4_PORT = 143
IMAP4_SSL_PORT = 993

def sendMail(*args,**kwargs):
    '''Enviar email,
       @Arguments:
           _sender       - str  -  enviado desde
           _to           - list -  enviar a
           _cc           - list -  enviar copias
           _bcc          - list -  enviar copias ocultas
           _subject      - str  -  asunto
           _text         - str  -  text/plain del mensaje
           _html         - str  -  text/html  del mensaje
           _user         - str  -  usuario
           _password     - str  -  password
           _smtpserver   - str  -  servidor SMTP
           _imapserver   - str  -  servidor IMAP
           _smtp_port    - str  -  puerto smtp
           _imap_port    - str  -  puerto imap
           _files        - list -  ficheros adjuntos
           _output       - str  -  el contenido del mensaje se grabara como un fichero
       @Returns 200|401
    '''
    # for standard SMTP protocol port 25, no encryption
    # for standard SMTP protocol port 465, SSL
    # for standard SMTP protocol port 587, TLS

    #Tratamiento de errores en LOG


    try:
        gc.collect()
        size = 0
        return_code = None
        return_text = ''
        valid_email = re.compile(r'^.+@.+')
        # Create the root message and fill in the from, to, and subject headers
        # keeps the textual and attachment parts separately
        msgRoot = MIMEMultipart('mixed')
        msgRoot['Subject'] = bytes2str(kwargs.get('_subject',''))
        # my email address
        msgRoot['From'] = bytes2str(kwargs.get('_sender',''))
        lstTo = list()
        
        if kwargs.get('_to',None) and valid_email.match(str(kwargs.get('_to',''))):
            # recipient's email address
            _to = kwargs.get('_to') if isinstance(kwargs.get('_to'),(list,tuple)) else bytes2str([kwargs.get('_to')]);
            msgRoot['To'] = COMMASPACE.join(_to)
            lstTo = _to
            if kwargs.get('_cc',None) and valid_email.match(str(kwargs.get('_cc',''))):
                _cc = kwargs.get('_cc') if isinstance(kwargs.get('_cc'),(list,tuple)) else bytes2str([kwargs.get('_cc')]);
                msgRoot['cc'] = COMMASPACE.join(_cc)
                lstTo += _cc
            if kwargs.get('_bcc',None) and valid_email.match(str(kwargs.get('_bcc',''))):
                _bcc = kwargs.get('_bcc') if isinstance(kwargs.get('_bcc'),(list,tuple)) else bytes2str([kwargs.get('_bcc')]);
                msgRoot['bcc'] = COMMASPACE.join(_bcc)
                lstTo += _bcc
            #msgRoot['Date'] = timeZone('Madrid').strftime( "%m/%d/%Y %H:%M" )
            msgRoot.preamble = 'This is a multi-part message in MIME format.'

            # Encapsulate the plain and HTML versions of the message body in an
            # 'alternative' part, so message agents can decide which they want to display.
            msgAlternative = MIMEMultipart('alternative')
            msgRoot.attach(msgAlternative)
            # MIMEText(_text[, _subtype[, _charset]])
            # Record the MIME types - text/plain
            msgAlternative.attach(MIMEText(kwargs.get('_text',''), _subtype='plain',_charset='utf-8'))
            # Record the MIME types - text/html.
            msgAlternative.attach(MIMEText(kwargs.get('_html',''), _subtype='html',_charset='utf-8'))

            # Ejemplo de como introducir una imagen en el texto html
            #'<b>Some <i>HTML</i> text</b> and an image.<br><img src="cid:image1"><br>Nifty!'

            # We reference the file in the SRC attribute by the ID
            if kwargs.get('_files',None):
                _files = kwargs['_files']
                for f in _files or []:
                    try:
                        if isfile(f):
                            size += os.path.getsize(f)
                            with open(f, "rb") as fr:
                                part = MIMEApplication(
                                    fr.read(),
                                    Name=basename(f)
                                )
                            # After the file is closed
                            part['Content-Disposition'] = 'attachment; filename="%s"' % basename(f)
                            msgRoot.attach(part)
                    except Exception as e:
                        tb = sys.exc_info()[2]
                        LOG.error('sendEmail <%s> in line %s !!!' % (repr(e),tb.tb_lineno))
                        del msgRoot
                        return 401

            # Send the email (this example assumes SMTP authentication is required)
            port = kwargs.get('_smtp_port',SMTP_DEFAULT_PORT)  # or 587
            if port==SMTP_SSL_PORT:                  # SSL
                server = SMTP_SSL(host=kwargs.get('_smtpserver',''),port=port,timeout=1000*6)
            else:
                server = SMTP(host=kwargs.get('_smtpserver',''),port=port,timeout=1000*6)
            # debuglevel
            # server.set_debuglevel(True) # show communication with the server

            # 220, b'pymetrick.com ESMTP Haraka/2.8.27 ready'
            return_code, return_text = server.connect(kwargs.get('_smtpserver',None), port)
            if return_code >= 200 and return_code <= 299 and server.ehlo()[0] == 250:
                #code = server.ehlo()[0]
                maximum_message_size_in_bytes = int( server.esmtp_features['size'] )
                LOG.debug('********** TAMAÑO MAXIMO ****************')
                LOG.debug(maximum_message_size_in_bytes)
                LOG.debug(f'''attachment size {size}''')
                if IS_SSL and server.has_extn('STARTTLS') and port==SMTP_TLS_PORT:
                    context = ssl.create_default_context()
                    server.starttls(context=context)
                    server.ehlo()   # re-identify ourselves over TLS connection
                if kwargs.get('_output',None):
                    with open(kwargs.get('_output','.'), 'w') as fp:
                        fp.write(msg.as_string())
                else:
                    # 235, b'2.7.0 Authentication successful'
                    return_code, return_text = server.login(kwargs.get('_user',None),kwargs.get('_password',None))
                    if return_code >= 200 and return_code <= 299:
                        server.sendmail(kwargs.get('_sender',''), lstTo, msgRoot.as_string())
                        # append a mail message to an IMAP SEND folder
                        port = kwargs.get('_imap_port',IMAP4_PORT)  # or 143
                        if port==IMAP4_SSL_PORT:                    # SSL
                            imap = IMAP4_SSL(host=kwargs.get('_imapserver',''),port=port)
                        else:
                            imap = IMAP4(host=kwargs.get('_imapserver',''),port=port)
                       
                        if IS_SSL and kwargs.get('_imap_port',None) and kwargs.get('_imapserver',None):
                            # OK, [b'Logged in']
                            return_code, return_text = imap.login(kwargs.get('_user',None),kwargs.get('_password',None))
                            imap.append('Sent', '', Time2Internaldate(time.time()), str(msgRoot).encode('utf8'))
                            imap.logout()
                        return return_code, return_text
                    else:
                        return return_code, return_text
                return return_code, return_text
            else:
                return return_code, return_text
        else:
            return 401,'ERROR - eMail erroneo !!!'
    except smtplib.SMTPException as e:
        tb = sys.exc_info()[2]
        LOG.error('SMTPException <%s> en linea %s !!!' % (repr(e),tb.tb_lineno))
        server.quit()
        return 401,repr(e)
    except smtplib.SMTPResponseException as e:
        tb = sys.exc_info()[2]
        LOG.error('SMTPResponseException <%s> en linea %s !!!' % (repr(e),tb.tb_lineno))
        server.quit()
        return 401,repr(e)
    except smtplib.SMTPSenderRefused as e:
        tb = sys.exc_info()[2]
        LOG.error('SMTPSenderRefused <%s> en linea %s !!!' % (repr(e),tb.tb_lineno))
        server.quit()
        return 401,repr(e)
    except smtplib.SMTPRecipientsRefused as e:
        tb = sys.exc_info()[2]
        LOG.error('SMTPRecipientsRefused <%s> en linea %s !!!' % (repr(e),tb.tb_lineno))
        server.quit()
        return 401,repr(e)
    except smtplib.SMTPDataError as e:
        tb = sys.exc_info()[2]
        LOG.error('SMTPDataError <%s> en linea %s !!!' % (repr(e),tb.tb_lineno))
        server.quit()
        return 401,repr(e)
    except smtplib.SMTPConnectError as e:
        tb = sys.exc_info()[2]
        LOG.error('SMTPConnectError <%s> en linea %s !!!' % (repr(e),tb.tb_lineno))
        server.quit()
        return 401,repr(e)
    except smtplib.SMTPHeloError as e:
        tb = sys.exc_info()[2]
        LOG.error('SMTPHeloError <%s> en linea %s !!!' % (repr(e),tb.tb_lineno))
        server.quit()
        return 401,repr(e)
    except smtplib.SMTPNotSupportedError as e:
        tb = sys.exc_info()[2]
        LOG.error('SMTPNotSupportedError <%s> en linea %s !!!' % (repr(e),tb.tb_lineno))
        server.quit()
        return 401,repr(e)
    except smtplib.SMTPAuthenticationError as e:
        tb = sys.exc_info()[2]
        LOG.error('SMTPAuthenticationError <%s> en linea %s !!!' % (repr(e),tb.tb_lineno))
        server.quit()
        return 401,repr(e)
    except smtplib.SMTPNotSupportedError as e:
        tb = sys.exc_info()[2]
        LOG.error('SMTPNotSupportedError <%s> en linea %s !!!' % (repr(e),tb.tb_lineno))
        server.quit()
        return 401,repr(e)
    except Exception as e:
        tb = sys.exc_info()[2]
        LOG.error('sendEmail <%s> in line %s !!!' % (repr(e),tb.tb_lineno))
        server.quit()
        return 401,repr(e)
    finally:
        del msgRoot
        del server
        del imap
        gc.collect()


def getMail(*args,**kwargs):
    '''Recibir email POP3 o IMAP
       @Arguments:
           _user         - str  -  usuario
           _password     - str  -  password
           _imapserver   - str  -  servidor IMAP
           _pop3server   - str  -  servidor POP3
           _port         - str  -  puerto
           _criteria     - str  -   ALL - devuelve todos los mensajes que coinciden con el resto del criterio
                                    ANSWERED - coincide con los mensajes con la bandera \\ANSWERED establecida
                                    BCC "cadena" - coincide con los mensajes con "cadena" en el campo Bcc:
                                    BEFORE "fecha" - coincide con los mensajes con Date: antes de "fecha"
                                    BODY "cadena" - coincide con los mensajes con "cadena" en el cuerpo del mensaje
                                    CC "cadena" - coincide con los mensajes con "cadena" en el campo Cc:
                                    DELETED - coincide con los mensajes borrados
                                    FLAGGED - coincide con los mensajes con la bandera \\FLAGGED establecida (algunas veces referidos como Importante o Urgente)
                                    FROM "cadena" - coincide con los mensajes con "cadena" en el campo From:
                                    KEYWORD "cadena" - coincide con los mensajes con "cadena" como palabra clave
                                    NEW - coincide con los mensajes nuevos
                                    OLD - coincide con los mensajes antiguos
                                    ON "fecha" - coincide con los mensajes con Date: coincidiendo con "fecha"
                                    RECENT - coincide con los mensajes con la bandera \\RECENT establecida
                                  * SEEN - coincide con los mensajes que han sido leídos (la bandera \\SEEN esta estabecido)
                                    SINCE "fecha" - coincide con los mensajes con Date: despues de "fecha"
                                    SUBJECT "cadena" - coincide con los mensajes con "cadena" en Subject:
                                    TEXT "cadena" - coincide con los mensajes con el texto "cadena"
                                    TO "cadena" - coincide con los mensajes con "cadena" en To:
                                    UNANSWERED - coincide con los mensajes que no han sido respondidos
                                    UNDELETED - coincide con los mensajes que no están eliminados
                                    UNFLAGGED - coincide con los mensajes que no tienen bandera
                                    UNKEYWORD "cadena" - coincide con los mensajes que no tienen la palabra clave "cadena"
                                  * UNSEEN - coincide con los mensajes que aun no han sido leidos
                         ejemplo cuando se incluye más de un criterio :  imap.search(None, 'UNSEEN', 'FROM "%s"' % (sender_of_interest))
                                  
           _outputdir    - str  -  fichero donde grabara los correos leidos
       @Returns 200|401

    '''
    # for standard IMAP protocol port 143, no encryption
    # for standard IMAP protocol port 993, SSL
    # for standard IMAP protocol port 143, TLS


    try:
        input_data = []

        if kwargs.get('_imapserver',None):
            # Receive the email (this example assumes IMAP authentication is required)
            port = kwargs.get('_port',IMAP4_PORT)

            if port==IMAP4_SSL_PORT:
                server = IMAP4_SSL(kwargs['_imapserver'],port)
            else:
                server = IMAP4(kwargs['_imapserver'],port)

            if IS_SSL and port == IMAP4_PORT: # situar en NOAUTH
                context = ssl.create_default_context()
                server.starttls(ssl_context=context)

            server.login(kwargs.get('_user',None),kwargs.get('_password',None))   # AUTH

            #server.select(mailbox='INBOX', readonly=False)
            server.select(readonly=False)
            typ, data = server.search(None, kwargs.get('_criteria','UNSEEN'))
            if typ == 'OK':
                for msgid in data[0].split():
                    typ, data = server.fetch(msgid, '(RFC822)')
                    email_message=email.message_from_bytes(data[0][1])
                    # Header Details
                    date_tuple = email.utils.parsedate_tz(email_message['Date'])
                    if date_tuple:
                        local_date = datetime.fromtimestamp(email.utils.mktime_tz(date_tuple))
                        local_message_date = "%s" %((local_date.strftime("%a, %d %b %Y %H:%M:%S")))
                    email_from = (email.header.make_header(email.header.decode_header(email_message['From'])))
                    email_to = (email.header.make_header(email.header.decode_header(email_message['To'])))
                    email_subject = (email.header.make_header(email.header.decode_header(email_message['Subject'])))
                    # Body details
                    body_data = {'body':'','file':[]}
                    for part in email_message.walk():
                        
                        if part.get_content_type() == "text/plain":
                            body = bytes2str(part.get_payload(decode=True),'utf-8')
                            print("From: %s\nTo: %s\nDate: %s\nSubject: %s\n\nBody: \n\n%s" %(email_from, email_to,local_message_date, email_subject, body))
                            file_name = "email_" + bytes2str(msgid) + ".txt"
                            body_data['body'] = file_name
                            output_file = open(kwargs.get('_outputdir','') + file_name, 'w')

                            output_file.write("From: %s\nTo: %s\nDate: %s\nSubject: %s\n\nBody: \n\n%s" %(email_from, email_to,local_message_date, email_subject, body))
                            output_file.close()
                        if part.get_content_maintype() != 'multipart' and part.get('Content-Disposition') is not None:
                            body_data['file'].append(part.get_filename())
                            open(kwargs.get('_outputdir','') + part.get_filename(), 'wb').write(part.get_payload(decode=True))
                            

                        else:
                            continue
                    input_data.append(body_data)
                server.close()
            server.logout()
            del server
            return 200, 'application/json', json.dumps(input_data,default=to_json)
        else:
            return 401
    except Exception as e:
        tb = sys.exc_info()[2]
        LOG.error('getEmail <%s> in line %s !!!' % (repr(e),tb.tb_lineno))
        print('getEmail <%s> in line %s !!!' % (repr(e),tb.tb_lineno))
        return 401

def deleteIMAP(*args,**kwargs):
    '''Eliminar todos los correos de IMAP
       @Arguments:
           _user         - str  -  usuario
           _password     - str  -  password
           _imapserver   - str  -  servidor IMAP
           _port         - str  -  puerto
           _folder       - str  -  carpeta a tratar  'INPUT',
           _criteria     - str  -   ALL - devuelve todos los mensajes que coinciden con el resto del criterio
                                    ANSWERED - coincide con los mensajes con la bandera \\ANSWERED establecida
                                    BCC "cadena" - coincide con los mensajes con "cadena" en el campo Bcc:
                                    BEFORE "fecha" - coincide con los mensajes con Date: antes de "fecha"
                                    BODY "cadena" - coincide con los mensajes con "cadena" en el cuerpo del mensaje
                                    CC "cadena" - coincide con los mensajes con "cadena" en el campo Cc:
                                    DELETED - coincide con los mensajes borrados
                                    FLAGGED - coincide con los mensajes con la bandera \\FLAGGED establecida (algunas veces referidos como Importante o Urgente)
                                    FROM "cadena" - coincide con los mensajes con "cadena" en el campo From:
                                    KEYWORD "cadena" - coincide con los mensajes con "cadena" como palabra clave
                                    NEW - coincide con los mensajes nuevos
                                    OLD - coincide con los mensajes antiguos
                                    ON "fecha" - coincide con los mensajes con Date: coincidiendo con "fecha"
                                    RECENT - coincide con los mensajes con la bandera \\RECENT establecida
                                  * SEEN - coincide con los mensajes que han sido leídos (la bandera \\SEEN esta estabecido)
                                    SINCE "fecha" - coincide con los mensajes con Date: despues de "fecha"
                                    SUBJECT "cadena" - coincide con los mensajes con "cadena" en Subject:
                                    TEXT "cadena" - coincide con los mensajes con el texto "cadena"
                                    TO "cadena" - coincide con los mensajes con "cadena" en To:
                                    UNANSWERED - coincide con los mensajes que no han sido respondidos
                                    UNDELETED - coincide con los mensajes que no están eliminados
                                    UNFLAGGED - coincide con los mensajes que no tienen bandera
                                    UNKEYWORD "cadena" - coincide con los mensajes que no tienen la palabra clave "cadena"
                                  * UNSEEN - coincide con los mensajes que aun no han sido leidos
                         ejemplo cuando se incluye más de un criterio :  imap.search(None, 'UNSEEN', 'FROM "%s"' % (sender_of_interest))           
       @Returns 200|401
    '''

    try:
        if kwargs.get('_imapserver',None) and kwargs.get('_user',None) and kwargs.get('_password',None) and kwargs.get('_folder',None):
            port = kwargs.get('_port',IMAP4_PORT)
            
            if port==IMAP4_SSL_PORT:
                server = IMAP4_SSL(kwargs['_imapserver'],port)
            else:
                server = IMAP4(kwargs['_imapserver'],port)

            if IS_SSL and port == IMAP4_PORT: # situar en NOAUTH
                context = ssl.create_default_context()
                server.starttls(ssl_context=context)
                
            server.login(kwargs.get('_user',None),kwargs.get('_password',None))   # AUTH
            if kwargs.get('_folder',None):
                server.select(mailbox='INBOX', readonly=False)
            else:
                server.select(readonly=False)
                
            #typ, data = server.search(None, 'ALL')  # para eliminar todos los emails
            typ, data = server.search(None,kwargs.get('_criteria','DELETED'))
            if typ == 'OK':
                for msgid in data[0].split():
                    server.store(msgid, '+FLAGS', '\\Deleted')
                    # server.store(msgid, '+FLAGS', '\\Trash')
            server.expunge()
            server.close()
            server.logout()
            del server
            return 200
        else:
            return 401
    except Exception as e:
        tb = sys.exc_info()[2]
        LOG.error('deleteIMAP <%s> in line %s !!!' % (repr(e),tb.tb_lineno))
        return 401

def timeZone(timeZone):
    """Adaptar la hora segun la zona horaria"""
    zone = {'Madrid':2,'Londres':1,'Kabul':4.5}
    nowUTC = datetime.utcnow()
    return nowUTC + timedelta(hours=zone[timeZone])

if __name__ == "__main__":
    print ('''copyright {0}'''.format( __copyright__))
    print ('''license {0}'''.format( __license__))
    print ('''version {0}'''.format( __version__))
    if len(sys.argv) < 2:
        sys.stderr.write("for help use -h o --help")
    elif sys.argv[1]=='-h' or sys.argv[1]=='--help':
        print ('''
        Podra enviar fácilmente mensajes de correo electronico :\n\n''')

    #getMail(_user='pymetrick@pymetrick.org',_password='BETACA96',_imapserver='pymetrick.com',_port=993,_criteria='''SUBJECT "[OK] - Let's encrypt certificate issued"''',_outputdir='c:/tmp/prueba.txt')
    #f = getMail(_user='pymetrick@pymetrick.org',_password='BETACA96',_imapserver='pymetrick.com',_port=993,_criteria='''SUBJECT "envio"''',_outputdir='c:/tmp/')
    #print(f)
    #deleteIMAP(_user='pymetrick@pymetrick.org',_password='BETACA96',_imapserver='pymetrick.com',_port=993,_folder='INBOX',_criteria='''SUBJECT "envio"''')






#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of Pymetrick.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.

"""Modulo para gestionar S3 BUCKETS de AWS o compatibles"""

try:
    from pymetrick.version import __author__, __version__, __license__, __copyright__
except ImportError:
    raise ImportError(
        'Failed to find (autogenerated) version.py '
    )

__date__ = '2012-09-21'
__credits__ = ''
__text__ = 'Tratamiento de S3 Buckets'
__file__ = 'bucket.py'
__boto3_version__ = '1.26.84'

#--- CHANGES ------------------------------------------------------------------
# 2021-09-14 v0.01 PL: - First version

import sys
# sin bytecode
sys.dont_write_bytecode = True

#from secrets import AWS_DEFAULT_REGION, AWS_ACCESS_KEY_ID, AWS_SECRET_ACCESS_KEY, endpoint_url

import botocore
import boto3
from boto3.s3.transfer import S3Transfer, TransferConfig

import os
import queue

import logging
import logging.handlers

try:
    import simplejson as json
except:
    import json

'''
DEBUG - debug message
INFO - info message
WARNING - warn message
ERROR - error message
CRITICAL - critical message
'''
LOG_LEVEL_VALUE = {'DEBUG':10,'INFO':20,'WARNING':30,'ERROR':40,'CRITICAL':50}
if str(os.environ.get('PYMETRICK_LOG_LEVEL',None)).upper() in LOG_LEVEL_VALUE:
    LOG_LEVEL = LOG_LEVEL_VALUE[str(os.environ.get('PYMETRICK_LOG_LEVEL')).upper()]
else:
    LOG_LEVEL = 30
LOG_FILENAME = '-'.join([os.path.abspath(__file__).split(os.sep)[len(os.path.abspath(__file__).split(os.sep))-1],])[:-3]
LOG = logging.getLogger(LOG_FILENAME)

if 'LD_LIBRARY_PATH' in list(os.environ.keys()):
    # CGI environment
    sys.stdout = sys.stderr
    logging.basicConfig(stream = sys.stderr, level=LOG_LEVEL, format='%(filename)s[line:%(lineno)d] - %(levelname)s - %(message)s')
else:
    # not CGI environment
    logging.basicConfig(stream=sys.stderr)
    hdlr = logging.handlers.RotatingFileHandler(filename=LOG_FILENAME+'.log',mode='a', encoding='utf-8', maxBytes=1048576, backupCount=3)
    formatter = logging.Formatter('%(asctime)s - %(filename)s[line:%(lineno)d] - %(levelname)s - %(message)s')
    hdlr.setFormatter(formatter)
    LOG.addHandler(hdlr)
    LOG.setLevel(LOG_LEVEL)
    
GB = 1024 ** 3

class Bucket(object):
    '''Gestion de S3 Buckets'''

    def __init__(self, *args,**kwargs):
        '''
        @parameters
          aws_access_key_id
          aws_secret_access_key
          endpoint_url
        '''
        try:
            # Define the configuration rules
            self.cors_configuration = {
                    'CORSRules': [{
                    'AllowedHeaders': ['Authorization'],
                    'AllowedMethods': ['GET', 'PUT', 'POST', 'DELETE'],
                    'AllowedOrigins': ['*'],
                    'ExposeHeaders': ['ETag', 'x-amz-request-id'],
                    'MaxAgeSeconds': 3000
                }]
            }

            if 'default_bucket' in kwargs:
                self.default_bucket = kwargs['default_bucket']
            else:
                self.default_bucket = None

            object_storage_config = {
            'aws_access_key_id' : None,
            'aws_secret_access_key' : None,
            'endpoint_url': None }

            for key in object_storage_config.keys():
                if key in kwargs:
                    object_storage_config[key] = kwargs[key]
                elif key == 'cors':
                    self.cors_configuration = cors if (isinstance(cors,(dict,)) and 'CORSRules' in cors) else None
                    
                
            self.client = boto3.client('s3', **object_storage_config )
            self.resource = boto3.resource('s3', **object_storage_config)

            if self.default_bucket and self.cors_configuration:
                self.client.put_bucket_cors(Bucket=self.default_bucket, CORSConfiguration=self.cors_configuration)

        except botocore.exceptions.ClientError as error:
            for key, value in sorted(botocore.exceptions.__dict__.items()):
                if isinstance(value, type):
                    print(key)
        except botocore.exceptions.ParamValidationError as error:
            raise ValueError('The parameters you provided are incorrect: {}'.format(error))


    def list_buckets(self):
        '''
        list buckets
        '''
        try:
            list_buckets = []
            response = self.client.list_buckets()
            if isinstance(response,dict) and "HTTPStatusCode" in response["ResponseMetadata"] and response["ResponseMetadata"]["HTTPStatusCode"]==200:
                if 'Buckets' in response:
                    for bucket in response['Buckets']:
                        list_buckets.append(bucket['Name'])
                return list_buckets
            else:
                return ['Error']

        except botocore.exceptions.ClientError as error:
            for key, value in sorted(botocore.exceptions.__dict__.items()):
                if isinstance(value, type):
                    print(key)
        except botocore.exceptions.ParamValidationError as error:
            raise ValueError('The parameters you provided are incorrect: {}'.format(error))

    def versioning(self, version=False):
        '''enabling|suspend versioning for s3 bucket object'''
        try:
            _status_init = False
            _status_final = False 
            if self.default_bucket:
                _versioning = self.resource.BucketVersioning(self.default_bucket)
                _status_init = _versioning.status

                if version:
                    _versioning.enable()
                else:
                    _versioning.suspend()

                _status_final = _versioning.status

                return [_status_init, _status_final]
            
        except botocore.exceptions.ClientError as error:
            for key, value in sorted(botocore.exceptions.__dict__.items()):
                if isinstance(value, type):
                    print(key)
        except botocore.exceptions.ParamValidationError as error:
            raise ValueError('The parameters you provided are incorrect: {}'.format(error))                    
            

    def list_objects(self, prefix='', maxkeys=1000):
        try:

            list_objects = []
            
            # 10000 max result line for request
            response = self.client.list_objects_v2(Bucket=self.default_bucket, MaxKeys=maxkeys, Prefix=prefix)
            if isinstance(response,dict) and "HTTPStatusCode" in response["ResponseMetadata"] and response["ResponseMetadata"]["HTTPStatusCode"]==200:
                if 'Contents' in response:
                    for key in response['Contents']:
                        list_objects.append(key['Key'])

                while response['IsTruncated'] and maxkeys>=1000:
                    continuation_key = response['NextContinuationToken']
                    response = self.client.list_objects_v2(Bucket=self.default_bucket, MaxKeys=maxkeys, Delimiter="/", ContinuationToken=continuation_key, Prefix=prefix)
                    if 'Contents' in response:
                        for key in response['Contents']:
                            list_objects.append(key['Key'])
                            if maxkeys != 1000:
                                if len(list_objects)==maxkeys:
                                    break
                    
                return list_objects
            else:
                return ['Error']
            

        except botocore.exceptions.ClientError as error:
            for key, value in sorted(botocore.exceptions.__dict__.items()):
                if isinstance(value, type):
                    print(key)
        except botocore.exceptions.ParamValidationError as error:
            raise ValueError('The parameters you provided are incorrect: {}'.format(error))


    def add_bucket(self, bucket_name=None ):
        try:
            if bucket_name:
                self.client.create_bucket(Bucket=bucket_name)
                self.default_bucket = bucket_name
        except botocore.exceptions.ClientError as error:
            for key, value in sorted(botocore.exceptions.__dict__.items()):
                if isinstance(value, type):
                    print(key)
        except botocore.exceptions.ParamValidationError as error:
            raise ValueError('The parameters you provided are incorrect: {}'.format(error))


    def delete_bucket(self, bucket_name=None ):
        try:
            if bucket_name:
                self.client.delete_bucket(Bucket=bucket_name)
            else:
                self.client.delete_bucket(Bucket=self.default_bucket)

        except botocore.exceptions.ClientError as error:
            for key, value in sorted(botocore.exceptions.__dict__.items()):
                if isinstance(value, type):
                    print(key)
        except botocore.exceptions.ParamValidationError as error:
            raise ValueError('The parameters you provided are incorrect: {}'.format(error))


    def upload_object(self, prefix='', list_objects=[], extra_args=None):
        ''' Upload a file to an S3 bucket
            @parameters
            prefix: Bucket path
            list_objects: S3 list object name
            extra_args: dict with 'extra_args' like extra_args={'ACL':'public-read'} or
            extra_args={'GrantRead': 'uri="http://acs.amazonaws.com/groups/global/AllUsers"',
                'GrantFullControl': 'id="01234567890abcdefg"',}
            extra_args={'ACL': 'public-read', 'ContentType': 'image/png'}

            @return
            :True if file was uploaded, else False
        '''
        try:
            transfer = S3Transfer(self.client)
            config = TransferConfig(multipart_threshold=5*GB)
            for file in list_objects:
                upload_file_key = (''.join([prefix,os.path.split(file)[-1:][0]])).upper()
                transfer.upload_file(
                                  filename=file,
                                  bucket=self.default_bucket,
                                  key=upload_file_key,
                                  extra_args=extra_args)
                
            return True
        except botocore.exceptions.ClientError as error:
            for key, value in sorted(botocore.exceptions.__dict__.items()):
                if isinstance(value, type):
                    LOG.error(f'''bucket.upload_object {key}''')
            return False
        except botocore.exceptions.ParamValidationError as error:
            LOG.error(ValueError('The parameters you provided are incorrect: {}'.format(error)))
            raise ValueError('The parameters you provided are incorrect: {}'.format(error))
            return False
        except Exception as e:
            if hasattr(e, 'message'):
                LOG.error('<%s> in line %s !!!' % (e.message,format(sys.exc_info()[-1].tb_lineno)))
            else:
                LOG.error('<%s> in line %s !!!' % (e,format(sys.exc_info()[-1].tb_lineno)))
            return False

    #from pathlib import Path
    def download_object(self, prefix='', list_objects=[]):
        try:
            objects = []
            transfer = S3Transfer(self.client)
            if len(list_objects)==0:
                list_objects = self.client.list_objects(prefix=prefix)
                prefix=None
            for file in list_objects:
                if prefix:
                    download_file_key = (''.join([prefix,os.path.split(file)[-1:][0]])).upper()
                else:
                    download_file_key = file.upper()
                path, filename = os.path.split(file.upper())
                #Create sub directories if its not existing
                if not os.path.exists(path):
                    try:
                        os.makedirs(path)
                    except OSError:
                        if not os.path.isdir(path):
                            raise
                    #Path(path).mkdir(parents=True, exist_ok=True)
                try:
                    transfer.download_file(
                        bucket=self.default_bucket,
                        key=download_file_key,
                        filename=os.sep.join([path,filename]))
                    objects.append(os.sep.join([path,filename]))
                except Exception as e:
                    print(e)
            return objects

        except botocore.exceptions.ClientError as error:
            for key, value in sorted(botocore.exceptions.__dict__.items()):
                if isinstance(value, type):
                    print(key)
        except botocore.exceptions.ParamValidationError as error:
            raise ValueError('The parameters you provided are incorrect: {}'.format(error))
        except Exception as e:
            print(e)


    def delete_object(self, prefix=None, list_objects=[]):
        try:
            if len(list_objects)==0:
                list_objects = self.client.list_objects(prefix=prefix)
                prefix=None                
            for file in list_objects:
                # delete object o directory
                if prefix:
                    delete_file_key = (''.join([prefix,os.path.split(file)[-1:][0]])).upper()
                else:
                    delete_file_key = file.upper()
                response = self.client.delete_object(
                    Bucket=self.default_bucket,
                    Key=delete_file_key)
                if 'DeleteMarker' in response:
                    return response['DeleteMarker']
                else:
                    return False
        except botocore.exceptions.ClientError as error:
            for key, value in sorted(botocore.exceptions.__dict__.items()):
                if isinstance(value, type):
                    print(key)
        except botocore.exceptions.ParamValidationError as error:
            raise ValueError('The parameters you provided are incorrect: {}'.format(error))

    def backup_object(self, prefix=''):
        try:
            _prefix = prefix
            _list_objects = self.list_objects(prefix=_prefix)
            return self.download_object(prefix=None,list_objects=_list_objects)
        except botocore.exceptions.ClientError as error:
            for key, value in sorted(botocore.exceptions.__dict__.items()):
                if isinstance(value, type):
                    print(key)
        except botocore.exceptions.ParamValidationError as error:
            raise ValueError('The parameters you provided are incorrect: {}'.format(error))        
        


if __name__  == "__main__":
    print ('''copyright {0}'''.format( __copyright__))
    print ('''license {0}'''.format( __license__))
    print ('''version {0}'''.format( __version__))
    if len(sys.argv) < 2:
        sys.stderr.write("for help use -h o --help")
    elif sys.argv[1]=='-h' or sys.argv[1]=='--help':
        print ('''
        Tratamiento de S3 Bucket\n\n''')





import datetime
import hashlib
import json
from flask import Flask, jsonify

class BlockChain:
    def __init__(self):
        self.chain = []
        self.create_block(proof=1,previous_hash='0')

    def create_block(self, proof, previous_hash):
        block = {'index':len(self.chain)+1,
                 'timestamp':str(datetime.datetime.now()),
                 'proof':proof,
                 'previous_hash':previous_hash}

        self.chain.append(block)
        return block

    def get_previous_block(self):
        return self.chain[-1]

    def proof_of_work(self,previous_proof):
        new_proof = 1
        check_proof = False

        while check_proof is False:
            hash_operation = hashlib.sha256(str(new_proof**2 - previous_proof**2).encode()).hexdigest()
            if hash_operation[:4] == '0000':
                print(hash_operation)
                check_proof = True
            else:
                new_proof += 1
        return new_proof


    def hash(self, block):
        encoded_block = json.dumps(block, sort_keys = True).encode()
        return hashlib.sha256(encoded_block).hexdigest()

    def is_chain_valid(self, chain):
        previous_block = chain[0]
        block_index = 1
        while block_index < len(chain):
            block = chain[block_index]
            if block['previous_hash'] != self.hash(previous_block):
                return False
            previous_proof = previous_block['proof']
            proof = block['proof']
            hash_operation = hashlib.sha256(str(proof**2-previous_proof**2).encode()).hexdigest()
            if hash_operation[:4] != '0000':
                return False

            previous_block = block
            block_index += 1
        return True


app = Flask(__name__)
blockchain = BlockChain()


@app.route('/')
def hello_world():
    return 'Hello, World!'

# minar bloques
@app.route('/mine_block',methods=['GET'])
def mine_block():
    previous_block = blockchain.get_previous_block()
    previous_proof = previous_block['proof']
    proof = blockchain.proof_of_work(previous_proof)
    previous_hash = blockchain.hash(previous_block)
    block = blockchain.create_block(proof,previous_hash)
    response = {'message':'Felicidades, has minado un bloque!',
                'index':block['index'],
                'timestamp':block['timestamp'],
                'proof':block['proof'],
                'previous_hash':block['previous_hash']}
    return jsonify(response), 200

# obtener todos los bloques
@app.route('/get_chain',methods=['GET'])
def get_chain():
    response = {'chain':blockchain.chain,
                'length':len(blockchain.chain)}
    return jsonify(response), 200

# chequear la validez de la cadena de bloques
@app.route('/is_valid',methods=['GET'])
def is_valid():
    is_valid = blockchain.is_chain_valid(blockchain.chain)
    if is_valid:
        response = {'message':'Todo bien. El Blockchain es valido'}
    else:
        response = {'message':'Hay un problema, Blockchanin no valido'}
    return jsonify(response), 200
 
app.run(host='0.0.0.0',port='5000')

       
            
                                    
                 



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# ****************************************************************************
# * Software: FPDF for python                                                *
# * Version:  1.7.1a                                                          *
# * Date:     2010-09-10                                                     *
# * Last update: 2019-01-01                                                  *
# * License:  LGPL v3.0                                                      *
# *                                                                          *
# * Original Author (PHP):  Olivier PLATHEY 2004-12-31                       *
# * Ported to Python 2.4 by Max (maxpat78@yahoo.it) on 2006-05               *
# * Maintainer:  Mariano Reingart (reingart@gmail.com) et al since 2008 est. *
# * NOTE: 'I' and 'D' destinations are disabled, and simply print to STDOUT  *
# * Updated only for PYMETRICK 01/01/2019 by javtamvi@pymetrick.org          * 
# ****************************************************************************

"Special module to handle differences between Python 2 and 3 versions"

import sys, os
# sin bytecode
sys.dont_write_bytecode = True

PY3K = sys.version_info >= (3, 0)

try:
    import cPickle as pickle
except ImportError:
    import pickle

try:
    from urllib import urlopen
except ImportError:
    from urllib.request import urlopen

try:
    from hashlib import md5
except ImportError:
    try:
        from md5 import md5
    except ImportError:
        md5 = None
def hashpath(fn):
    h = md5()
    if PY3K:
        h.update(fn.encode("UTF-8"))
    else:
        h.update(fn)
    return h.hexdigest()

# Check if PIL is available (tries importing both pypi version and corrected or manually installed versions).
# Necessary for JPEG and GIF support.
# TODO: Pillow support
try:
    from PIL import Image
except ImportError:
    try:
        import Image
    except ImportError:
        Image = None

try:
	from HTMLParser import HTMLParser
except ImportError:
	from html.parser import HTMLParser

if PY3K:
    basestring = str
    unicode = str
    ord = lambda x: x
else:
    basestring = basestring
    unicode = unicode
    ord = ord

import logging
import logging.handlers

'''
DEBUG - debug message
INFO - info message
WARNING - warn message
ERROR - error message
CRITICAL - critical message
'''
LOG_LEVEL_VALUE = {'DEBUG':10,'INFO':20,'WARNING':30,'ERROR':40,'CRITICAL':50}
if str(os.environ.get('PYMETRICK_LOG_LEVEL',None)).upper() in LOG_LEVEL_VALUE:
    LOG_LEVEL = LOG_LEVEL_VALUE[str(os.environ.get('PYMETRICK_LOG_LEVEL')).upper()]
else:
    LOG_LEVEL = 30
LOG_FILENAME = '-'.join([os.path.abspath(__file__).split(os.sep)[len(os.path.abspath(__file__).split(os.sep))-1],])[:-3]
LOG = logging.getLogger(LOG_FILENAME)    

# shortcut to bytes conversion (b prefix)
def b(s):
    if isinstance(s, basestring):
        return s.encode("latin9")
    elif isinstance(s, int):
        if PY3K:
            return bytes([s])       # http://bugs.python.org/issue4588
        else:
            return chr(s)
                      
def exception():
    "Return the current the exception instance currently being handled"
    # this is needed to support Python 2.5 that lacks "as" syntax
    return sys.exc_info()[1]





#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of Pymetrick.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.

"""Modulo para gestionar identificacion fiscal y monedas"""

try:
    from pymetrick.version import __author__, __version__, __license__, __copyright__
except ImportError:
    raise ImportError(
        'Failed to find (autogenerated) version.py '
    )

__date__ = '2012-09-21'
__credits__ = ''
__text__ = 'Tratamiento fiscal, bancos y monedas'
__file__ = 'exchequer.py'

#--- CHANGES ------------------------------------------------------------------
# 2012-09-21 v0.01 PL: - First version

import sys
# sin bytecode
sys.dont_write_bytecode = True
import os
import string
import re
import urllib.request
import ssl
import zeep
import logging
import logging.handlers

try:
    import simplejson as json
except:
    import json

'''
DEBUG - debug message
INFO - info message
WARNING - warn message
ERROR - error message
CRITICAL - critical message
'''
LOG_LEVEL_VALUE = {'DEBUG':10,'INFO':20,'WARNING':30,'ERROR':40,'CRITICAL':50}
if str(os.environ.get('PYMETRICK_LOG_LEVEL',None)).upper() in LOG_LEVEL_VALUE:
    LOG_LEVEL = LOG_LEVEL_VALUE[str(os.environ.get('PYMETRICK_LOG_LEVEL')).upper()]
else:
    LOG_LEVEL = 30
LOG_FILENAME = '-'.join([os.path.abspath(__file__).split(os.sep)[len(os.path.abspath(__file__).split(os.sep))-1],])[:-3]
LOG = logging.getLogger(LOG_FILENAME)

if 'LD_LIBRARY_PATH' in list(os.environ.keys()):
    # CGI environment
    sys.stdout = sys.stderr
    logging.basicConfig(stream = sys.stderr, level=LOG_LEVEL, format='%(filename)s[line:%(lineno)d] - %(levelname)s - %(message)s')
else:
    # not CGI environment
    logging.basicConfig(stream=sys.stderr)
    hdlr = logging.handlers.RotatingFileHandler(filename=LOG_FILENAME+'.log',mode='a', encoding='utf-8', maxBytes=1048576, backupCount=3)
    formatter = logging.Formatter('%(asctime)s - %(filename)s[line:%(lineno)d] - %(levelname)s - %(message)s')
    hdlr.setFormatter(formatter)
    LOG.addHandler(hdlr)
    LOG.setLevel(LOG_LEVEL)

'''
País	Tarifa cero?	Tarifa reducida	Tarifa estándar
Austria	      No           10% o 12%	20%
Bélgica	      Sí	   6% o 12%	21%
Bulgaria	No	   9%	        20%
República Checa	No	   15%	        21%
Croacia	        No	  5% o 13%	25%
Chipre	        Sí	  5% o 9%	19%
Dinamarca	Sí	     -	        25%
Estonia	        Sí	     9%	        20%
Finlandia	Sí	10% o 14%	24%
Francia	        No	2.1% o 5.5% o 10%	20%
Alemania	No	7%	19%
Grecia	        No	6.5% o 13%	23%
Hungría	        No	5% o 18%	27%
Irlanda	        Sí	4.8% o 9% o 13.5%	23%
Italia	        Sí	4% o 10%	22%
Letonia	        No	12%	        21%
Lituania	No	5% o 9%	        21%
Luxemburgo	No	3% o 8%	        17%
Malta	        Sí	5% o 7%	        18%
Países Bajos	No	6%	        21%
Polonia	        Sí	5% o 8%	        23%
Portugal	No	6% o 13%	23%
Rumania	        No	5% o 9%	        24%
Eslovenia	No	9.5%	        22%
Eslovaquia	No	10%	        20%
España	        No	4% o 10%	21%
Suecia	        Sí	6% o 12%	25%
Reino Unido	Sí	5%	        20%
'''

ISO_COUNTRY_2_TAX_LIST = ['AT','BE','BG','CY','CZ','DE','DK','EE','EL','ES','FI','FR','HR','HU','IE','IT','LT','LU','LV','MT','NL','PL','PT','RO','GB','SE','SI','SK']

def validate_EORI(iso_country='ES',nif=''):
    '''Tax license validation
       @parameters
           iso_country: ISO country. ES(default)
           nif:   tax license
       @return
           iso_country.nif  <-- if it's right
           &iso_country.nif <-- if it's wrong
    '''
    try:
        iso_country_2 = ''.join(re.findall("^[a-zA-Z]{2}",nif)).upper() if ''.join(re.findall("^[a-zA-Z]{2}",nif)).upper() in ISO_COUNTRY_2_TAX_LIST else iso_country
        # elimina las siglas de iso_country_2 en nif 
        _nif = " ".join(re.findall("[a-zA-Z0-9]+",nif)).upper().replace(' ','')[len(''.join(re.findall("^[a-zA-Z]{2}",nif))):]
        if iso_country_2=='AT':
            # AUSTRIA ATU99999999 = AT + 1 bloque con una "U" seguida de 8 dígitos numéricos p.e.:ATU12345675
            # FORMATO
            #          A(1)+N(1-7)+C(1)
            #          A(1)={U}
            # S=N(1)+2#N(2)+N(3)+2#N(4)+N(5)+2#N(6)+N(7)
            # C(1)=10-(S+4)%10; si C(1)=10, C(1)=0
            return (iso_country_2,''.join(re.findall("^[a-zA-Z]{2}",nif)).upper()+_nif)
        elif iso_country_2=='BE':
            # BELGICA BE9999999999 = BE + 1 bloque de 10 dígitos numéricos, el primero "0" o "1". p.e.:BE1234567894
            # FORMATO
            #         N(1-8)+C(1-2)
            #         N(2)>1
            # C(1-2)=97-N(1-7)%97
            # 
            return (iso_country_2,''.join(re.findall("^[a-zA-Z]{2}",nif)).upper()+_nif)
        elif iso_country_2=='BG':
            # BULGARIA BG9999999999 = BG + 1 bloque de 9 o 10 dígitos numéricos
            __nif = ''.join(re.findall("[0-9]{9,10}",_nif))
            if len(__nif) == 9 or len(__nif) == 10 :
                return (iso_country_2,''.join(re.findall("^[a-zA-Z]{2}",nif)).upper()+_nif)
            # identificacion fiscal CZ incorrecta
            return (iso_country_2,'&'+''.join(re.findall("^[a-zA-Z]{2}",nif)).upper()+_nif)
        elif iso_country_2=='CY':
            # CHIPRE CY99999999L = CY + 1 bloque de 8 dígitos numéricos y al final 1 letra
            return (iso_country_2,''.join(re.findall("^[a-zA-Z]{2}",nif)).upper()+_nif)
        elif iso_country_2=='CZ':
            # CHEQUIA CZ9999999999 = CZ + 1 bloque de 8, 9 o 10 dígitos numéricos
            __nif = ''.join(re.findall("[0-9]{8,10}",_nif))
            if len(__nif) == 8 or len(__nif) == 9 or len(__nif) == 10 :
                return (iso_country_2,''.join(re.findall("^[a-zA-Z]{2}",nif)).upper()+_nif)
            # identificacion fiscal CZ incorrecta
            return (iso_country_2,'&'+''.join(re.findall("^[a-zA-Z]{2}",nif)).upper()+_nif)
        elif iso_country_2=='DE':
            # ALEMANIA DE999999999 = DE + 1 bloque de 9 dígitos numéricos
            __nif = ''.join(re.findall("[0-9]{9}",_nif))
            if len(__nif) == 9 :
                return (iso_country_2,''.join(re.findall("^[a-zA-Z]{2}",nif)).upper()+_nif)
            # identificacion fiscal DE incorrecta
            return (iso_country_2,'&'+''.join(re.findall("^[a-zA-Z]{2}",nif)).upper()+_nif)
        elif iso_country_2=='DK':
            # DINAMARCA DK99 99 99 99 = DK + 4 bloques de 2 dígitos numéricos
            __nif = ''.join(re.findall("[0-9]{8}",_nif))
            if len(__nif) == 8 :
                return (iso_country_2,''.join(re.findall("^[a-zA-Z]{2}",nif)).upper()+_nif)
            # identificacion fiscal DK incorrecta
            return (iso_country_2,'&'+''.join(re.findall("^[a-zA-Z]{2}",nif)).upper()+_nif)
        elif iso_country_2=='EE':
            # ESTONIA EE999999999 = EE + 1 bloque de 9 dígitos numéricos
            __nif = ''.join(re.findall("[0-9]{9}",_nif))
            if len(__nif) == 9 :
                return (iso_country_2,''.join(re.findall("^[a-zA-Z]{2}",nif)).upper()+_nif)
            # identificacion fiscal EE incorrecta
            return (iso_country_2,'&'+''.join(re.findall("^[a-zA-Z]{2}",nif)).upper()+_nif)
        elif iso_country_2=='EL':
            # GRECIA EL999999999 = EL + 1 bloque de 9 dígitos numéricos
            __nif = ''.join(re.findall("[0-9]{9}",_nif))
            if len(__nif) == 9 :
                return (iso_country_2,''.join(re.findall("^[a-zA-Z]{2}",nif)).upper()+_nif)
            # identificacion fiscal EL incorrecta
            return (iso_country_2,'&'+''.join(re.findall("^[a-zA-Z]{2}",nif)).upper()+_nif)
        elif iso_country_2=='ES':
            # ESX9999999X = ES + 1 bloque de 9 caracteres (primero y último: letra o dígitos, resto: sólo dígitos).
            _nif = " ".join(re.findall("[a-zA-Z0-9]+",nif)).upper().replace(' ','')
            if _nif[:2] == iso_country_2:
                _nif = _nif[2:]
            ''' NIF 8 digitos + 1 caracter o 1 caracter + 7 digitos + 1 caracter
                CIF 7 digitos + 1 caracter de control
            '''
            ALPHA_NIF = 'TRWAGMYFPDXBNJZSQVHLCKE'
            '''Entidades juridicas con animo de lucro'''
            ALPHA_CIF_a = 'JABCDEFGHI'
            '''Personas juridicas sin animo de lucro '''
            ALPHA_CIF_b = 'NPQRSW'
            '''Personas fisicas extranjeras, menores de 14 o no residentes'''
            ALPHA_CIF_c = 'KLMXYZ'
            '''Comprobar si es un NIF'''
            if len(_nif)>0:
                if _nif[0] in '0123456789':
                    _nif_number = ' '.join(re.findall("[0-9]+",_nif))
                    ctrl = (int(_nif_number))%23
                    __nif = _nif_number+ALPHA_NIF[ctrl]
                    if __nif.strip().upper() == _nif:
                        __nif = _nif_number.zfill(8)+ALPHA_NIF[ctrl]
                        return (iso_country_2,__nif)
                else:
                    '''Comprobar si es una Persona fisica extranjera, menor de 14 o no residente'''
                    if _nif[0] in ALPHA_CIF_c:
                        _nif_number = ' '.join(re.findall("[0-9]+",_nif))
                        '''Si contiene 7 digitos y comienza por X Y Z, sustituir X Y Z por 0 1 2'''
                        ctrl = int(_nif_number)
                        if len(_nif_number) == 7 and len(_nif) == 9:
                            if _nif[0] in 'XYZ':
                                n = {'X':'0','Y':'1','Z':'2'}
                                ctrl = int(n[_nif[0]]+_nif_number)
                        ctrl = ctrl%23
                        __nif = _nif[0]+_nif_number+ALPHA_NIF[ctrl]
                        if __nif.strip().upper() == _nif:
                            __nif = _nif[0]+_nif_number.zfill(7)+ALPHA_NIF[ctrl]
                            return (iso_country_2,__nif)
                    '''Comprobar si es una Persona juridica sin animo de lucro'''
                    if _nif[0] in ALPHA_CIF_a:
                        if 7<len(_nif)<=9:
                            _nif_number = ' '.join(re.findall("[0-9]+",_nif))
                            '''obtener posicion par e impar'''
                            n = 0
                            n_ctrl = 0
                            for n in range(0,len(_nif_number)-1):
                                if (n+1)%2 == 0:
                                    n_ctrl += int(_nif_number[n])
                                else:
                                    n_ctrl += (int(_nif_number[n])*2%10)+int(int(_nif_number[n])*2/10)
                            ctrl = n_ctrl%10
                            ctrl = 10 - ctrl
                            if ctrl>9:
                                ctrl = 0
                            __nif = _nif[0]+_nif_number[0:7]+repr(ctrl)
                            if __nif.strip().upper() == _nif:
                                __nif = _nif[0]+_nif_number[0:7].zfill(7)+repr(ctrl)
                                return (iso_country_2,__nif)
                    '''Comprobar si es una Persona juridica sin animo de lucro'''
                    if _nif[0] in ALPHA_CIF_b:
                        if 7<len(_nif)<=9:
                            _nif_number = ' '.join(re.findall("[0-9]+",_nif))
                            '''obtener posicion par e impar'''
                            n = 0
                            n_ctrl = 0
                            for n in range(0,len(_nif_number)):
                                if (n+1)%2 == 0:
                                    n_ctrl += int(_nif_number[n])
                                else:
                                    n_ctrl += (int(_nif_number[n])*2%10)+int(int(_nif_number[n])*2/10)
                            ctrl = n_ctrl%10
                            ctrl = 10 - ctrl
                            __nif = _nif[0]+_nif_number[0:7]+chr(ctrl+64).upper()
                            if __nif.strip().upper() == _nif:
                                __nif = (_nif[0]+_nif_number[0:7].zfill(7)+chr(ctrl+64)).upper()
                                return (iso_country_2,__nif)
            # identificacion fiscal ES incorrecta
            return (iso_country_2,'&'+_nif)
        elif iso_country_2=='FI':
            # FINLANDIA FI99999999 = FI + 1 bloque de 8 dígitos numéricos
            __nif = ''.join(re.findall("[0-9]{8}",_nif))
            if len(__nif) == 8 :
                return (iso_country_2,''.join(re.findall("^[a-zA-Z]{2}",nif)).upper()+_nif)
            # identificacion fiscal FI incorrecta
            return (iso_country_2,'&'+''.join(re.findall("^[a-zA-Z]{2}",nif)).upper()+_nif)
        elif iso_country_2=='FR':
            # FRANCIA FRXX 999999999 = FR + 1 bloque de 2 caracteres y otro de 9 dígitos numéricos 
            # FORMATO  p.e. 32123456789
            # Sistema antiguo	C(1-2)+N(1-9)
            #                   N(1-9)>0	
            # C(1-2)=(N(1-9)*100+12)%97
            # Sistema nuevo p.e. 2H123456789
            #                   C(1-2)+N(1-9)
            #                   N(1-9)>0
            #                   C(1-2)={A-H,J-N,P-Z} | {0-9}
            #                   si C(1) es alfabético, C(2) ha de ser numérico y al revés.	
            #                   Para un N(1-9) dado, hay más de un C(1-2) posible, por tanto, no se puede calcular éste.
            # C{0-9,A-H,J-N,P-Z} → C{0-33}
            # si C(1)<10, S=C(1)*24+C(2)-10
            # si C(1)>9, S=C(1)*34+C(2)-100
            # X=S%11
            # S=S\11+1
            # Y=(N(1-9)+S)%11
            # si X=Y, correcto
            return (iso_country_2,''.join(re.findall("^[a-zA-Z]{2}",nif)).upper()+_nif)
        elif iso_country_2=='HR':
            # CROACIA
            return (iso_country_2,''.join(re.findall("^[a-zA-Z]{2}",nif)).upper()+_nif)
        elif iso_country_2=='HU':
            # HUNGRIA HU99999999 = HU + 1 bloque de 8 dígitos numéricos
            __nif = ''.join(re.findall("[0-9]{8}",_nif))
            if len(__nif) == 8 :
                return (iso_country_2,''.join(re.findall("^[a-zA-Z]{2}",nif)).upper()+_nif)
            # identificacion fiscal HU incorrecta
            return (iso_country_2,'&'+''.join(re.findall("^[a-zA-Z]{2}",nif)).upper()+_nif)
        elif iso_country_2=='IE':
            # IRLANDA IE9S99999L = IE + 1 bloque de 8 dígitos. "S" puede ser letra (A-Z), dígito (0-9); "+" o "*".
            return (iso_country_2,''.join(re.findall("^[a-zA-Z]{2}",nif)).upper()+_nif)
        elif iso_country_2=='IT':
            # ITALIA IT99999999999 = IT + 1 bloque de 11 dígitos numéricos
            __nif = ''.join(re.findall("[0-9]{11}",_nif))
            if len(__nif) == 11 :
                return (iso_country_2,''.join(re.findall("^[a-zA-Z]{2}",nif)).upper()+_nif)
            # identificacion fiscal IT incorrecta
            return (iso_country_2,'&'+''.join(re.findall("^[a-zA-Z]{2}",nif)).upper()+_nif)
        elif iso_country_2=='LT':
            # LITUANIA LT999999999999 = LT + 1 bloque de 9 o 12 dígitos numéricos
            __nif = ''.join(re.findall("[0-9]{9,12}",_nif))
            if len(__nif) == 9 or len(__nif) == 12 :
                return (iso_country_2,''.join(re.findall("^[a-zA-Z]{2}",nif)).upper()+_nif)
            # identificacion fiscal LT incorrecta
            return (iso_country_2,'&'+''.join(re.findall("^[a-zA-Z]{2}",nif)).upper()+_nif)
        elif iso_country_2=='LU':
            # LUXEMBURGO LU99999999 = LU + 1 bloque de 8 dígitos numéricos
            __nif = ''.join(re.findall("[0-9]{8}",_nif))
            if len(__nif) == 8 :
                return (iso_country_2,''.join(re.findall("^[a-zA-Z]{2}",nif)).upper()+_nif)
            # identificacion fiscal LU incorrecta
            return (iso_country_2,'&'+''.join(re.findall("^[a-zA-Z]{2}",nif)).upper()+_nif)
        elif iso_country_2=='LV':
            # LETONIA LV99999999999 = LV + 1 bloque de 11 dígitos numéricos
            __nif = ''.join(re.findall("[0-9]{11}",_nif))
            if len(__nif) == 11 :
                return (iso_country_2,''.join(re.findall("^[a-zA-Z]{2}",nif)).upper()+_nif)
            # identificacion fiscal LV incorrecta
            return (iso_country_2,'&'+''.join(re.findall("^[a-zA-Z]{2}",nif)).upper()+_nif)
        elif iso_country_2=='MT':
            # MALTA MT99999999 = MT + 1 bloque de 8 dígitos numéricos
            __nif = ''.join(re.findall("[0-9]{8}",_nif))
            if len(__nif) == 8 :
                return (iso_country_2,''.join(re.findall("^[a-zA-Z]{2}",nif)).upper()+_nif)
            # identificacion fiscal MT incorrecta
            return (iso_country_2,'&'+''.join(re.findall("^[a-zA-Z]{2}",nif)).upper()+_nif)
        elif iso_country_2=='NL':
            # PAISES BAJOS NL999999999B99 = NL + 1 bloque de 12 caracteres, siendo el décimo siempre "B".
            return (iso_country_2,''.join(re.findall("^[a-zA-Z]{2}",nif)).upper()+_nif)
        elif iso_country_2=='PL':
            # POLONIA PL9999999999 = PL + 1 bloque de 10 dígitos numéricos
            __nif = ''.join(re.findall("[0-9]{10}",_nif))
            if len(__nif) == 10 :
                return (iso_country_2,''.join(re.findall("^[a-zA-Z]{2}",nif)).upper()+_nif)
            # identificacion fiscal PL incorrecta
            return (iso_country_2,'&'+''.join(re.findall("^[a-zA-Z]{2}",nif)).upper()+_nif)
        elif iso_country_2=='PT':
            # PORTUGAL PT999999999 = PT + 1 bloque de 9 dígitos numéricos p.e.:123456789
            # FORMATO
            #          N(1-8)+C(1)
            #          N(1)>0
            # S=9*N(1)+8*N(2)+7*N(3)+6*N(4)+5*N(5)+4*N(6)+3*N(7)+2*N(8)
            # C(1)=11-S%11; si C(1)=10, C(1)=0; si C(1)=11, C(1)=0
            if len(_nif)==9:
                n = [ int(x) for x in _nif ]
                s = n[0]*9+n[1]*8+n[2]*7+n[3]*6+n[4]*5+n[5]*4+n[6]*3+n[7]*2
                c = 11-s%11
                dc = '0' if c == 10 or c == 11 else str(c)
                __nif = _nif[:8]+dc
                if __nif.strip().upper() == _nif:
                    return (iso_country_2,''.join(re.findall("^[a-zA-Z]{2}",nif)).upper()+__nif)
            # identificacion fiscal PT incorrecta
            return (iso_country_2,'&'+''.join(re.findall("^[a-zA-Z]{2}",nif)).upper()+_nif)
        elif iso_country_2=='RO':
            # RUMANIA RO9999999999 = RO + 1 bloque de 2 a 10 dígitos numéricos
            __nif = ''.join(re.findall("[0-9]{2,10}",_nif))
            if len(__nif) >= 2 and len(__nif) <= 10 :
                return (iso_country_2,''.join(re.findall("^[a-zA-Z]{2}",nif)).upper()+_nif)
            # identificacion fiscal RO incorrecta
            return (iso_country_2,'&'+''.join(re.findall("^[a-zA-Z]{2}",nif)).upper()+_nif)
        elif iso_country_2=='GB':
            # REINO UNIDO GB999 9999 99 999 = GR + 3 bloques de 3, 4 y 2 dígitos numéricos.
            #             Empresas de grupos: + otro de 3 dígitos.  /  Dep. .gubernamentales: GBGD999  /  Autoridades sanitarias: GBHA999
            return (iso_country_2,''.join(re.findall("^[a-zA-Z]{2}",nif)).upper()+_nif)
        elif iso_country_2=='SE':
            # SUECIA SE999999999901 = 1 bloque de 12 dígitos numéricos, los dos últimos siempre "01"
            __nif = ''.join(re.findall("[0-9]{12}",_nif))
            if len(__nif) == 12 and __nif[10:]=='01' :
                return (iso_country_2,''.join(re.findall("^[a-zA-Z]{2}",nif)).upper()+_nif)
            # identificacion fiscal SE incorrecta
            return (iso_country_2,'&'+''.join(re.findall("^[a-zA-Z]{2}",nif)).upper()+_nif)
        elif iso_country_2=='SI':
            # ESLOVENIA SI99999999 = SI + 1 bloque de 8 dígitos numéricos
            __nif = ''.join(re.findall("[0-9]{8}",_nif))
            if len(__nif) == 8:
                return (iso_country_2,''.join(re.findall("^[a-zA-Z]{2}",nif)).upper()+_nif)
            # identificacion fiscal SI incorrecta
            return (iso_country_2,'&'+''.join(re.findall("^[a-zA-Z]{2}",nif)).upper()+_nif)
        elif iso_country_2=='SK':
            # ESLOVAQUIA SK9999999999 = SK + 1 bloque de 10 dígitos numéricos
            __nif = ''.join(re.findall("[0-9]{10}",_nif))
            if len(__nif) == 10:
                return (iso_country_2,''.join(re.findall("^[a-zA-Z]{2}",nif)).upper()+_nif)
            # identificacion fiscal SK incorrecta
            return (iso_country_2,'&'+''.join(re.findall("^[a-zA-Z]{2}",nif)).upper()+_nif)
        else:
            # identificacion fiscal ajena
            return ('00',_nif)

        # identificacion fiscal NO comprobada
        return (iso_country_2,_nif)
    except Exception as e:
        if hasattr(e, 'message'):
            LOG.error('<%s> in line %s !!!' % (e.message,format(sys.exc_info()[-1].tb_lineno)))
        else:
            LOG.error('<%s> in line %s !!!' % (e,format(sys.exc_info()[-1].tb_lineno)))

#******************************************* Consulta de situación de contribuyentes en el registro de operadores intracomunitarios. ******************

VAT_NUMBER_REGEXPS = {
    'AT': re.compile(r'^U\d{8}$', re.IGNORECASE),
    'BE': re.compile(r'^\d{9,10}$'),
    'BG': re.compile(r'^\d{9,10}$'),
    'CY': re.compile(r'^\d{8}[a-z]$', re.IGNORECASE),
    'CZ': re.compile(r'^\d{8,10}$'),
    'DE': re.compile(r'^\d{9}$'),
    'DK': re.compile(r'^\d{8}$'),
    'EE': re.compile(r'^\d{9}$'),
    'EL': re.compile(r'^\d{9}$'),
    'ES': re.compile(r'^[\da-z]\d{7}[\da-z]$', re.IGNORECASE),
    'FI': re.compile(r'^\d{8}$'),
    'FR': re.compile(r'^[\da-hj-np-z]{2}\d{9}$', re.IGNORECASE),
    'GB': re.compile(r'^((\d{9})|(\d{12})|(GD\d{3})|(HA\d{3}))$', re.IGNORECASE),
    'HR': re.compile(r'^\d{11}$'),
    'HU': re.compile(r'^\d{8}$'),
    'IE': re.compile(r'^((\d{7}[a-z])|(\d[a-z]\d{5}[a-z])|(\d{6,7}[a-z]{2}))$', re.IGNORECASE),
    'IT': re.compile(r'^\d{11}$'),
    'LT': re.compile(r'^((\d{9})|(\d{12}))$'),
    'LU': re.compile(r'^\d{8}$'),
    'LV': re.compile(r'^\d{11}$'),
    'MT': re.compile(r'^\d{8}$'),
    'NL': re.compile(r'^\d{9}B\d{2,3}$', re.IGNORECASE),
    'PL': re.compile(r'^\d{10}$'),
    'PT': re.compile(r'^\d{9}$'),
    'RO': re.compile(r'^\d{2,10}$'),
    'SE': re.compile(r'^\d{12}$'),
    'SI': re.compile(r'^\d{8}$'),
    'SK': re.compile(r'^\d{10}$'),
}



#client = Client('https://ec.europa.eu/taxation_customs/vies/checkVatTestService.wsdl')

class ViesValidationError(Exception):
    """
    Exception thrown by the Vies object when there is a
    validation problem.
    """
    pass


class ViesHTTPError(Exception):
    """
    Exception thrown by the Vies object when there is an
    HTTP error interacting with Vies.
    """
    pass


class ViesError(Exception):
    """
    Base Exception thrown by the Vies object when there is a
    general problem when interacting with Vies service.
    """
    pass

class Vies(object):
    '''Consulta de situación de contribuyentes en el registro de operadores intracomunitarios.
       ejemplo:
       v = Vies()
       v.request(vat_number='PTddddddddd')
       v.request(vat_number='Bdddddddd',vat_country_code='ES')
       v.request(vat_number='ESBdddddddd',vat_country_code='ES')

    @parameteres
       vat_country: ISO country. ES(default)
       vat_number : tax license

    @return
        {
        'countryCode': 'PT',
        'vatNumber': 'ddddddddd',
        'requestDate': datetime.date(2022, 4, 3),
        >'valid': True|False, <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
        'traderName': 'NOMBRE SUPUESTO',
        'traderCompanyType': '---',
        'traderAddress': 'DIRECCION COMPLETA SUPUESTA',
        'traderStreet': None,
        'traderPostcode': None,
        'traderCity': None,
        'traderNameMatch': None,
        'traderCompanyTypeMatch': None,
        'traderStreetMatch': None,
        'traderPostcodeMatch': None,
        'traderCityMatch': None,
        'requestIdentifier': None
        }

    '''

    WS_ENDPOINT = 'http://ec.europa.eu/taxation_customs/vies/checkVatService.wsdl'
    
    def __init__(self):
        super(Vies, self).__init__()

    def request(self, vat_number, vat_country_code=None, extended_info=True):
        try:
            vat_number = vat_number.replace(' ','')
            
            if vat_country_code is not None:
                vat_country_code = vat_country_code.replace(' ','').upper()

            else:
                vat_country_code = vat_number[:2]

            if vat_country_code not in VAT_NUMBER_REGEXPS.keys():
                vat_country_code = 'ES'

            if vat_number[:2].upper() == vat_country_code:
                vat_number = vat_number[2:]    

            if vat_country_code=='EL':
                vat_country_code = 'GR'

            if vat_country_code not in VAT_NUMBER_REGEXPS.keys():
                raise ViesValidationError('Invalid VAT country')

            # check VIES
            try:
                client = zeep.Client(wsdl=self.WS_ENDPOINT)
                client.settings(raw_response=True, force_https=True)
                
            except Exception as e:
                raise ViesHTTPError('VIES service unreachable. %s' % str(e))

            try:
                if extended_info is True:
                    result = client.service.checkVatApprox(vat_country_code, vat_number)
                else:
                    result = client.service.checkVat(vat_country_code, vat_number)
            except Exception as e:
                raise ViesError('Got error from vies: %s' % str(e))


            return result            
                
        except Exception as e:
            if hasattr(e, 'message'):
                LOG.error('<%s> in line %s !!!' % (e.message,format(sys.exc_info()[-1].tb_lineno)))
            else:
                LOG.error('<%s> in line %s !!!' % (e,format(sys.exc_info()[-1].tb_lineno)))
                return {}


#******************************************* End Consulta de situación de contribuyentes en el registro de operadores intracomunitarios. ******************

def currencyExchangeEUR(dest=None):
    '''Get daily list of changes from ECB in XML format
       and extract currency exchange information from 1 EUR
       @parameters
           dest: ISO currency
       @return
           [changes,symbol]  <-- if it's right
           [0,'']            <-- if it's wrong
    '''
    try:
        currency_list = list()
        currency_dict = dict()
        url = "https://www.ecb.europa.eu/stats/eurofxref/eurofxref-daily.xml"
        if (not os.environ.get('PYTHONHTTPSVERIFY', '') and getattr(ssl, '_create_unverified_context', None)): 
            ssl._create_default_https_context = ssl._create_unverified_context
        currency = urllib.request.urlopen(url, context=ssl.SSLContext()).read()
        buscar = b"currency\s\'[\w{3}]\s\'?rate=\'(\d{4,7}.)\'/>"
        l_currency = re.findall(b"currency='[\w]{3}'\srate='[\d]{1,5}[.]?[\d]{1,5}'",currency)
        for m in l_currency:
            currency_list = list(m.decode().replace('currency=','').replace(' rate','').replace("'",'').split('='))
            currency_dict[currency_list[0]]=float(currency_list[1])
        if dest is not None and dest in currency_dict:
            symbol = currencyCode(code=dest)[-1]
            return [currency_dict[dest],symbol]
        else:
            return [0,'']
    except Exception as e:
        if hasattr(e, 'message'):
            LOG.error('<%s> in line %s !!!' % (e.message,format(sys.exc_info()[-1].tb_lineno)))
        else:
            LOG.error('<%s> in line %s !!!' % (e,format(sys.exc_info()[-1].tb_lineno)))


def currencyCode(code='EUR'):
    '''Check currency
       @parameters
           code: ISO currency
       @return
           [ISO currency,symbol]  <-- if it's right
           ['','']                <-- if it's wrong
    '''
    try:
        _code =''
        # return [ISO, symbol]
        _code_list = ['','']
        currency_dict = {'USD':'$',
                        'EUR':'eur|€',
                        'JPY':'yen|円|¥',
                        'BGN':'lev|leva|лев|лева',
                        'CZK':'koruna česká|Kč',
                        'DKK':'krone|kroner|kr',
                        'GBP':'pound sterling|£',
                        'HUF':'forint',
                        'PLN':'esloti|złoty|zł',
                        'RON':'leu|lei',
                        'SEK':'krona|kronor',
                        'CHF':'franken|franc|franco',
                        'ISK':'króna|krónur|kr',
                        'NOK':'krone|kroner|kr',
                        'HRK':'kuna',
                        'RUB':'rublo|рубль|₽',
                        'TRY':'lira|Türk lirası',
                        'AUD':'australian dollar',
                        'BRL':'real|reais|R$',
                        'CAD':'canadian dollar|dollar canadien',
                        'CNY':'renminbi|RMB|yuan|人民币|人民幣|¥',
                        'HKD':'港元|港幣|HK$',
                        'IDR':'rupia|rupiah|Rp',
                        'ILS':'שקל חדש|₪',
                        'INR':'rupia|रुपया| روپیہ|₹',
                        'KRW':'대한민국 원|大韓民國 圓|daehan minguk weon|taehan minkuk wŏn|원|圓|weon|wŏn|₩',
                        'MXN':'peso mexicano|$',
                        'MYR':'ringgit|ريڠݢيت مليسيا|RM',
                        'NZD':'new zealand dollar|NZ$',
                        'PHP':'piso|php|₱',
                        'SGD':'singaporean dollar|元|வெள்ளி|ringgit|新加坡元|சிங்கப்பூர் வெள்ளி|S$',
                        'THB':'baht|บาท|฿',
                        'ZAR':'rand|R'
                       }
        if code.upper() in currency_dict.keys():
            _code = code
        else:
            for n in currency_dict.keys():
                currency_list = {x.lower() if type(x).__name__=='str' else x for x in currency_dict[n].split('|')}
                if code in currency_list:
                    _code = n
        if len(_code)>0 and _code in currency_dict.keys():
            # [ISO, symbol]
            _code_list = [_code,currency_dict[_code].split('|')[-1]]
        return _code_list
    except Exception as e:
        if hasattr(e, 'message'):
            LOG.error('<%s> in line %s !!!' % (e.message,format(sys.exc_info()[-1].tb_lineno)))
        else:
            LOG.error('<%s> in line %s !!!' % (e,format(sys.exc_info()[-1].tb_lineno)))

if __name__ == '__main__' :
    """
    print ('''copyright {0}'''.format( __copyright__))
    print ('''license {0}'''.format( __license__))
    print ('''version {0}'''.format( __version__))
    if len(sys.argv) < 2:
        sys.stderr.write("for help use -h o --help")
    elif sys.argv[1]=='-h' or sys.argv[1]=='--help':
        print ('''
        Tratamiento de identificaciones fiscales\n\n
        y gestion de monedas  ''')
    """




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""" Set of ESC/POS Commands (Constants)

This module contains constants that are described in the esc/pos-documentation.
Since there is no definitive and unified specification for all esc/pos-like printers the constants could later be
moved to `capabilities` as in `escpos-php by @mike42 <https://github.com/mike42/escpos-php>`_.

:author: `Manuel F Martinez <manpaz@bashlinux.com>`_ and others
:organization: Bashlinux and `python-escpos <https://github.com/python-escpos>`_
:copyright: Copyright (c) 2012-2017 Bashlinux and python-escpos
:license: MIT
"""
import sys
# sin bytecode
sys.dont_write_bytecode = True

import six

# Control characters
# as labelled in https://www.novopos.ch/client/EPSON/TM-T20/TM-T20_eng_qr.pdf
NUL = b'\x00'
EOT = b'\x04'
ENQ = b'\x05'
DLE = b'\x10'
DC4 = b'\x14'
CAN = b'\x18'
ESC = b'\x1b'
FS  = b'\x1c'
GS  = b'\x1d'

# Currency with CODEPAGE CP1252 p.e.: CODEPAGE_CHANGE + b'\x10'    default CODEPAGE CP437  CODEPAGE_CHANGE + b'\x00'   
CURRENCY = { 'euro'   : b'\x80',
             'dollar' : b'\x24',
             'yen'    : b'\xa5',
             'yuan'   : b'\xa5',
             'pound'  : b'\xa3'}

# Feed control sequences
CTL_LF = b'\x0a'              # Print and line feed
CTL_FF = b'\x0c'              # Form feed
CTL_CR = b'\x0d'              # Carriage return
CTL_HT = b'\x09'              # Horizontal tab
CTL_SET_HT = ESC + b'\x44'  # Set horizontal tab positions
CTL_VT = b'\x0b'              # Vertical tab

# Printer hardware
HW_INIT   = ESC + b'@'             # Clear data in buffer and reset modes
HW_SELECT = ESC + b'=\x01'         # Printer select

HW_RESET  = ESC + b'\x3f\x0a\x00'   # Reset printer hardware
                                    # (TODO: Where is this specified?)

# Cash Drawer (ESC p <pin> <on time: 2*ms> <off time: 2*ms>)
_CASH_DRAWER = lambda m, t1='', t2='': ESC + b'p' + m + six.int2byte(t1) + six.int2byte(t2)
CD_KICK_DEC_SEQUENCE = lambda esc, p, m, t1=50, t2=50: six.int2byte(esc) + six.int2byte(p) + six.int2byte(m) + six.int2byte(t1) + six.int2byte(t2)
CD_KICK_2 = _CASH_DRAWER(b'\x00', 50, 50)  # Sends a pulse to pin 2 []
CD_KICK_5 = _CASH_DRAWER(b'\x01', 50, 50)  # Sends a pulse to pin 5 []

# Paper Cutter
_CUT_PAPER = lambda m: GS + b'V' + m
PAPER_FULL_CUT = _CUT_PAPER(b'\x00')  # Full cut paper
PAPER_PART_CUT = _CUT_PAPER(b'\x01')  # Partial cut paper

# Beep (please note that the actual beep sequence may differ between devices)
BEEP = b'\x07'

# Panel buttons (e.g. the FEED button)
_PANEL_BUTTON = lambda n: ESC + b'c5' + six.int2byte(n)
PANEL_BUTTON_ON = _PANEL_BUTTON(0)  # enable all panel buttons
PANEL_BUTTON_OFF = _PANEL_BUTTON(1)  # disable all panel buttons

# Line display printing
LINE_DISPLAY_OPEN = ESC + b'\x3d\x02'
LINE_DISPLAY_CLEAR = ESC + b'\x40'
LINE_DISPLAY_CLOSE = ESC + b'\x3d\x01'

# Sheet modes
SHEET_SLIP_MODE = ESC + b'\x63\x30\x04'  # slip paper
SHEET_ROLL_MODE = ESC + b'\x63\x30\x01'  # paper roll

# Text format
# TODO: Acquire the "ESC/POS Application Programming Guide for Paper Roll
#       Printers" and tidy up this stuff too.
TXT_SIZE       = GS + b'!'

TXT_NORMAL     = ESC + b'!\x00'     # Normal text


TXT_STYLE = {
    'bold': {
        False: ESC + b'\x45\x00',               # Bold font OFF
        True: ESC + b'\x45\x01'                 # Bold font ON
    },
    'underline': {
        0: ESC + b'\x2d\x00',                   # Underline font OFF
        1: ESC + b'\x2d\x01',                   # Underline font 1-dot ON
        2: ESC + b'\x2d\x02'                    # Underline font 2-dot ON
    },
    'size': {
        'normal': TXT_NORMAL + ESC + b'!\x00',  # Normal text
        '2h': TXT_NORMAL + ESC + b'!\x10',      # Double height text
        '2w': TXT_NORMAL + ESC + b'!\x20',      # Double width text
        '2x': TXT_NORMAL + ESC + b'!\x30'       # Quad area text
    },
    'font': {
        'a': ESC + b'\x4d\x00',                 # Font type A
        'b': ESC + b'\x4d\x00'                  # Font type B
    },
    'align': {
        'left': ESC + b'\x61\x00',              # Left justification
        'center': ESC + b'\x61\x01',            # Centering
        'right': ESC + b'\x61\x02'              # Right justification
    },
    'invert': {
        True: GS  + b'\x42\x01',                # Inverse Printing ON
        False: GS  + b'\x42\x00'                # Inverse Printing OFF
    },
    'color': {
        'black': ESC + b'\x72\x00',             # Default Color
        'red': ESC + b'\x72\x01'                # Alternative Color, Usually Red
    },
    'flip': {
        True: ESC + b'\x7b\x01',                # Flip ON
        False: ESC + b'\x7b\x00'                # Flip OFF
    },
    'density': {
        0: GS + b'\x7c\x00',                    # Printing Density -50%
        1: GS + b'\x7c\x01',                    # Printing Density -37.5%
        2: GS + b'\x7c\x02',                    # Printing Density -25%
        3: GS + b'\x7c\x03',                    # Printing Density -12.5%
        4: GS + b'\x7c\x04',                    # Printing Density  0%
        5: GS + b'\x7c\x08',                    # Printing Density +50%
        6: GS + b'\x7c\x07',                    # Printing Density +37.5%
        7: GS + b'\x7c\x06',                    # Printing Density +25%
        8: GS + b'\x7c\x05'                     # Printing Density +12.5%
    },
    'smooth': {
        True: GS + b'\x62\x01',                 # Smooth ON
        False: GS + b'\x62\x00'                 # Smooth OFF
    },
    'height': {                                 # Custom text height
        1: 0x00,
        2: 0x01,
        3: 0x02,
        4: 0x03,
        5: 0x04,
        6: 0x05,
        7: 0x06,
        8: 0x07
    },
    'width': {                                  # Custom text width
        1: 0x00,
        2: 0x10,
        3: 0x20,
        4: 0x30,
        5: 0x40,
        6: 0x50,
        7: 0x60,
        8: 0x70
    }
}

# Fonts
SET_FONT = lambda n: ESC + b'\x4d' + n
TXT_FONT_A     = SET_FONT(b'\x00')  # Font type A
TXT_FONT_B     = SET_FONT(b'\x01')  # Font type B

# Spacing
LINESPACING_RESET = ESC + b'2'
LINESPACING_FUNCS = {
  60: ESC + b'A',  # line_spacing/60 of an inch, 0 <= line_spacing <= 85
  360: ESC + b'+', # line_spacing/360 of an inch, 0 <= line_spacing <= 255
  180: ESC + b'3', # line_spacing/180 of an inch, 0 <= line_spacing <= 255
}

# Prefix to change the codepage. You need to attach a byte to indicate
# the codepage to use. Transform DEC number to HEX number.
CODEPAGE_CHANGE = ESC + b'\x74'

'''
n Character Type Remark (size)
0 Page 0 USA, Standard Europe [CP437]
1 Page 1 Katakana
2 Page 2 Multilingual(Latin-1) [CP850]
3 Page 3 Portuguese [CP860]
4 Page 4 Canadian-French [CP863]
5 Page 5 Nordic [CP865]
6 Page 6 Slavic(Latin-2) [CP852]
7 Page 7 Turkish [CP857]
8 Page 8 Greek [CP737]
9 Page 9 Russian(Cyrillic) [CP866]
10 Page 10 Hebrew [CP862]
11 Page 11 Baltic [CP775]
12 Page 12 Polish
13 Page 13 Latin-9 [ISO8859-15]
14 Page 14 Latin1[Win1252]
15 Page 15 Multilingual Latin I + Euro[CP858]
16 Page 16 Russian(Cyrillic)[CP855]
17 Page 17 Russian(Cyrillic)[Win1251]
18 Page 18 Central Europe[Win1250]
19 Page 19 Greek[Win1253]
20 Page 20 Turkish[Win1254]
21 Page 21 Hebrew[Win1255]
22 Page 22 Vietnam[Win1258]
23 Page 23 Baltic[Win1257]
24 Page 24 Azerbaijani
25-29      Reserved
30         Thai[CP874]
31-39      Reserved
40 Page 25 Arabic [CP720]
41 Page 26 Arabic [Win 1256]
42 Page 27 Arabic (Farsi)
43 Page 28 Arabic presentation forms B
44-49      Reserved
50 Page 29 Hindi_Devanagari
252 Page 30 Japanese [CP932]
253 Page 31 Korean [CP949]
254 Page 32 Traditional Chinese [CP950]
255 Page 33 Simplified Chinese [CP936]
'''

# Prefix to change the international character set. You need to attach a
# byte to indicate the international character set to use.
INTERNATIONAL_CHARACTER_SET = ESC + b'\x52'
'''
n          Country          0x20 x20 x40 x50 x50 x50 x50 x60 x70 x70 x70 x7E
0          America           #    $   @   [   \   ]   ^   `   {   |   }   ~
1          France            #    $   à   °   ç   §   ^   `   é   ù   è   ̈
2          Germany           #    $   §   Ä   Ö   Ü   ^   `   ä   ö   ü   ß
3          UK                £    $   @   [   \   ]   ^   `   {   |   }   ~
4          Denmark I         #    $   @   Æ   Ø   Å   ^   `   æ   ø   å   ~
5          Sweden            #    ¤   É   Ä   Ö   Å   Ü   é   ä   ö   å   ü
6          Italy             #    $   @   °   \   é   ^   ù   à   ò   è   ì
7          Spain   I         ₧    $   @   ¡   Ñ   ¿   ^   ́    ̈   ñ   }   ~
8          Japan             #    $   @   [   ¥   ]   ^   `   {   |   }   ~
9          Norway            #    ¤   É   Æ   Ø   Å   Ü   é   æ   ø   å   ü
10         Denmark II        #    $   É   Æ   Ø   Å   Ü   é   æ   ø   å   ü
11         Spain   II        #    $   á   ¡   Ñ   ¿   é   ́    í   ñ   ó   ú
12         Latin Amer        #    $   @   ¡   Ñ   ¿   é   ü   í   ñ   ó   ú
13         Korea             #    $   @   [   ₩   ]   ^   `   {   |   }   ~          
'''

# Barcode format
_SET_BARCODE_TXT_POS = lambda n: GS + b'H' + n
BARCODE_TXT_OFF = _SET_BARCODE_TXT_POS(b'\x00')  # HRI barcode chars OFF
BARCODE_TXT_ABV = _SET_BARCODE_TXT_POS(b'\x01')  # HRI barcode chars above
BARCODE_TXT_BLW = _SET_BARCODE_TXT_POS(b'\x02')  # HRI barcode chars below
BARCODE_TXT_BTH = _SET_BARCODE_TXT_POS(b'\x03')  # HRI both above and below

_SET_HRI_FONT = lambda n: GS + b'f' + n
BARCODE_FONT_A = _SET_HRI_FONT(b'\x00')  # Font type A for HRI barcode chars
BARCODE_FONT_B = _SET_HRI_FONT(b'\x01')  # Font type B for HRI barcode chars

BARCODE_HEIGHT = GS + b'h'  # Barcode Height [1-255]
BARCODE_WIDTH  = GS + b'w'  # Barcode Width  [2-6]

# NOTE: This isn't actually an ESC/POS command. It's the common prefix to the
#      two "print bar code" commands:
#      -  Type A: "GS k <type as integer> <data> NUL"
#      -  TYPE B: "GS k <type as letter> <data length> <data>"
#      The latter command supports more barcode types
_SET_BARCODE_TYPE = lambda m: GS + b'k' + six.int2byte(m)

# Barcodes for printing function type A
BARCODE_TYPE_A = {
    'UPC-A':   _SET_BARCODE_TYPE(0),
    'UPC-E':   _SET_BARCODE_TYPE(1),
    'EAN13':   _SET_BARCODE_TYPE(2),
    'EAN8':    _SET_BARCODE_TYPE(3),
    'CODE39':  _SET_BARCODE_TYPE(4),
    'ITF':     _SET_BARCODE_TYPE(5),
    'NW7':     _SET_BARCODE_TYPE(6),
    'CODABAR': _SET_BARCODE_TYPE(6),  # Same as NW7
}

# Barcodes for printing function type B
# The first 8 are the same barcodes as type A
BARCODE_TYPE_B = {
    'UPC-A':                       _SET_BARCODE_TYPE(65),
    'UPC-E':                       _SET_BARCODE_TYPE(66),
    'EAN13':                       _SET_BARCODE_TYPE(67),
    'EAN8':                        _SET_BARCODE_TYPE(68),
    'CODE39':                      _SET_BARCODE_TYPE(69),
    'ITF':                         _SET_BARCODE_TYPE(70),
    'NW7':                         _SET_BARCODE_TYPE(71),
    'CODABAR':                     _SET_BARCODE_TYPE(71),  # Same as NW7
    'CODE93':                      _SET_BARCODE_TYPE(72),
    'CODE128':                     _SET_BARCODE_TYPE(73),
    'GS1-128':                     _SET_BARCODE_TYPE(74),
    'GS1 DATABAR OMNIDIRECTIONAL': _SET_BARCODE_TYPE(75),
    'GS1 DATABAR TRUNCATED':       _SET_BARCODE_TYPE(76),
    'GS1 DATABAR LIMITED':         _SET_BARCODE_TYPE(77),
    'GS1 DATABAR EXPANDED':        _SET_BARCODE_TYPE(78),
}

BARCODE_FORMATS = {
    'UPC-A':                       ([(11, 12)], "^[0-9]{11,12}$"),
    'UPC-E':                       ([(7, 8), (11, 12)], "^([0-9]{7,8}|[0-9]{11,12})$"),
    'EAN13':                       ([(12, 13)], "^[0-9]{12,13}$"),
    'EAN8':                        ([(7, 8)], "^[0-9]{7,8}$"),
    'CODE39':                      ([(1, 255)], "^([0-9A-Z \$\%\+\-\.\/]+|\*[0-9A-Z \$\%\+\-\.\/]+\*)$"),
    'ITF':                         ([(2, 255)], "^([0-9]{2})+$"),
    'NW7':                         ([(1, 255)], "^[A-Da-d][0-9\$\+\-\.\/\:]+[A-Da-d]$"),
    'CODABAR':                     ([(1, 255)], "^[A-Da-d][0-9\$\+\-\.\/\:]+[A-Da-d]$"),  # Same as NW7
    'CODE93':                      ([(1, 255)], "^[\\x00-\\x7F]+$"),
    'CODE128':                     ([(2, 255)], "^\{[A-C][\\x00-\\x7F]+$"),
    'GS1-128':                     ([(2, 255)], "^\{[A-C][\\x00-\\x7F]+$"), # same as CODE128
    'GS1 DATABAR OMNIDIRECTIONAL': ([(13,13)], "^[0-9]{13}$"),
    'GS1 DATABAR TRUNCATED':       ([(13,13)], "^[0-9]{13}$"), # same as GS1 omnidirectional
    'GS1 DATABAR LIMITED':         ([(13,13)], "^[01][0-9]{12}$"),
    'GS1 DATABAR EXPANDED':        ([(2,255)], "^\([0-9][A-Za-z0-9 \!\"\%\&\'\(\)\*\+\,\-\.\/\:\;\<\=\>\?\_\{]+$"),
}

BARCODE_TYPES = {
    'A': BARCODE_TYPE_A,
    'B': BARCODE_TYPE_B,
}

# QRCode error correction levels
QR_ECLEVEL_L = 0
QR_ECLEVEL_M = 1
QR_ECLEVEL_Q = 2
QR_ECLEVEL_H = 3

# QRcode models
QR_MODEL_1 = 1
QR_MODEL_2 = 2
QR_MICRO = 3

# Image format
# NOTE: _PRINT_RASTER_IMG is the obsolete ESC/POS "print raster bit image"
#       command. The constants include a fragment of the data's header.
_PRINT_RASTER_IMG = lambda data: GS + b'v0' + data
S_RASTER_N  = _PRINT_RASTER_IMG(b'\x00')  # Set raster image normal size
S_RASTER_2W = _PRINT_RASTER_IMG(b'\x01')  # Set raster image double width
S_RASTER_2H = _PRINT_RASTER_IMG(b'\x02')  # Set raster image double height
S_RASTER_Q  = _PRINT_RASTER_IMG(b'\x03')  # Set raster image quadruple

# Status Command
RT_STATUS = DLE + EOT
RT_STATUS_ONLINE = RT_STATUS +  b'\x01'
RT_STATUS_PAPER = RT_STATUS +  b'\x04'
RT_MASK_ONLINE = 8
RT_MASK_PAPER = 18
RT_MASK_LOWPAPER = 30
RT_MASK_NOPAPER = 114



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of Pymetrick.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.

"""Modulo para gestionar WSGI"""

try:
    from pymetrick.version import __author__, __version__, __license__, __copyright__
except ImportError:
    raise ImportError(
        'Failed to find (autogenerated) version.py '
    )
  
__date__ = '2014-05-21'
__credits__ = ''
__text__ = 'Gestion de funciones para WSGI'
__file__ = 'wsgi.py'

#--- CHANGES ------------------------------------------------------------------
# 2012-09-21 v0.01 PL: - First version
# 2017-10-02 v0.43 PL: - Migrar a PYTHON 3.6

import sys
# sin bytecode
sys.dont_write_bytecode = True
import os
import re
import base64
import logging
import logging.handlers
import collections
try:
  import simplejson as json
except:
  import json

from io import BytesIO, TextIOWrapper, BufferedReader

from pymetrick.helpers import *
from pymetrick.session import *


long = int

CONTENT_TYPE = {"7z":"application/x-7z-compressed",
                "aac":"audio/aac",
                "abw":"application/x-abiword",
                "arc":"application/octet-stream",
                "avi":"video/x-msvideo",
                "azw":"application/vnd.amazon.ebook",
                "bin":"application/octet-stream",
                "bmp":"image/bmp",
                "bz":"application/x-bzip",
                "bz2":"application/x-bzip2",
                "csh":"application/x-csh",
                "css":"text/css",
                "csv":"text/csv",
                "doc":"application/msword",
                "epub":"application/epub+zip",
                "gif":"image/gif",
                "htm":"text/html",
                "html":"text/html",
                "ico":"image/x-icon",
                "ics":"text/calendar",
                "jar":"application/java-archive",
                "jpeg":"image/jpeg",
                "jpg":"image/jpeg",
                "js":"application/javascript",
                "ical":"text/calendar",
                "ics":"text/calendar",
                "ifb":"text/calendar",
                "icalendar":"text/calendar",
                "json":"application/json",
                "mid":"audio/midi",
                "midi":"audio/midi",
                "mp4":"video/mp4",
                "mpeg":"video/mpeg",
                "mpkg":"application/vnd.apple.installer+xml",
                "odp":"application/vnd.oasis.opendocument.presentation",
                "ods":"application/vnd.oasis.opendocument.spreadsheet",
                "odt":"application/vnd.oasis.opendocument.text",
                "oga":"audio/ogg",
                "ogg":"audio/ogg",
                "ogv":"video/ogg",
                "ogx":"application/ogg",
                "pdf":"application/pdf",
                "png":"image/png",
                "ppt":"application/vnd.ms-powerpoint",
                "rar":"application/x-rar-compressed",
                "rtf":"application/rtf",
                "sh":"application/x-sh",
                "svg":"image/svg+xml",
                "swf":"application/x-shockwave-flash",
                "tar":"application/x-tar",
                "tif":"image/tiff",
                "tiff":"image/tiff",
                "ttf":"font/ttf",
                "txt":"text/plain",
                "vsd":"application/vnd.visio",
                "wav":"audio/x-wav",
                "weba":"audio/webm",
                "webm":"video/webm",
                "webp":"image/webp",
                "woff":"font/woff",
                "woff2":"font/woff2",
                "xhtml":"application/xhtml+xml",
                "xls":"application/vnd.ms-excel",
                "xml":"application/xml",
                "xul":"application/vnd.mozilla.xul+xml",
                "zip":"application/zip"}


STATUS_CODE = {
    "100": {
        "text": "Continue",
        "description": "\"indicates that the initial part of a request has been received and has not yet been rejected by the server.\"",
    },
    "101" : {
        "text": "Switching Protocols",
        "description": "\"indicates that the server understands and is willing to comply with the client's request, via the Upgrade header field, for a change in the application protocol being used on this connection.\"",
    },
    "102" : {
        "text": "Processing",
        "description": "\"is an interim response used to inform the client that the server has accepted the complete request, but has not yet completed it.\"",
    },
    "200" : {
        "text": "OK",
        "description": "\"indicates that the request has succeeded.\"",
    },
    "201" : {
        "text": "Created",
        "description": "\"indicates that the request has been fulfilled and has resulted in one or more new resources being created.\"",
    },
    "202" : {
        "text": "Accepted",
        "description": "\"indicates that the request has been accepted for processing, but the processing has not been completed.\"",
    },
    "203" : {
        "text": "Non-Authoritative Information",
        "description": "\"indicates that the request was successful but the enclosed payload has been modified from that of the origin server's 200 (OK) response by a transforming proxy.\"",
    },
    "204" : {
        "text": "No Content",
        "description": "\"indicates that the server has successfully fulfilled the request and that there is no additional content to send in the response payload body.\"",
    },
    "205" : {
        "text": "Reset Content",
        "description": "\"indicates that the server has fulfilled the request and desires that the user agent reset the \"document view\", which caused the request to be sent, to its original state as received from the origin server.\"",
    },
    "206" : {
        "text": "Partial Content",
        "description": "\"indicates that the server is successfully fulfilling a range request for the target resource by transferring one or more parts of the selected representation that correspond to the satisfiable ranges found in the requests's range header field.\"",
    },
    "207" : {
        "text": "Multi-Status",
        "description": "\"provides status for multiple independent operations.\"",
    },
    "226" : {
        "text": "IM Used",
        "description": "\"The server has fulfilled a GET request for the resource, and the response is a representation of the result of one or more instance-manipulations applied to the current instance.\"",
    },
    "300" : {
        "text": "Multiple Choices",
        "description": "\"indicates that the target resource has more than one representation, each with its own more specific identifier, and information about the alternatives is being provided so that the user (or user agent) can select a preferred representation by redirecting its request to one or more of those identifiers.\"",
    },
    "301" : {
        "text": "Moved Permanently",
        "description": "\"indicates that the target resource has been assigned a new permanent URI and any future references to this resource ought to use one of the enclosed URIs.\"",
    },
    "302" : {
        "text": "Found",
        "description": "\"indicates that the target resource resides temporarily under a different URI.\"",
    },
    "303" : {
        "text": "See Other",
        "description": "\"indicates that the server is redirecting the user agent to a different resource, as indicated by a URI in the Location header field, that is intended to provide an indirect response to the original request.\"",
    },
    "304" : {
        "text": "Not Modified",
        "description": "\"indicates that a conditional GET request has been received and would have resulted in a 200 (OK) response if it were not for the fact that the condition has evaluated to false.\"",
    },
    "305" : {
        "text": "Use Proxy",
        "description": "*deprecated*",
    },
    "307" : {
        "text": "Temporary Redirect",
        "description": "\"indicates that the target resource resides temporarily under a different URI and the user agent MUST NOT change the request method if it performs an automatic redirection to that URI.\"",
    },
    "308" : {
        "text": "Permanent Redirect",
        "description": "\"The target resource has been assigned a new permanent URI and any future references to this resource outght to use one of the enclosed URIs. [...] This status code is similar to 301 Moved Permanently (Section 7.3.2 of rfc7231), except that it does not allow rewriting the request method from POST to GET.\"",
    },
    "400" : {
        "text": "Bad Request",
        "description": "\"indicates that the server cannot or will not process the request because the received syntax is invalid, nonsensical, or exceeds some limitation on what the server is willing to process.\"",
    },
    "401" : {
        "text": "Unauthorized",
        "description": "\"indicates that the request has not been applied because it lacks valid authentication credentials for the target resource.\"",
    },
    "402" : {
        "text": "Payment Required",
        "description": "*reserved*",
    },
    "403" : {
        "text": "Forbidden",
        "description": "\"indicates that the server understood the request but refuses to authorize it.\"",
    },
    "404" : {
        "text": "Not Found",
        "description": "\"indicates that the origin server did not find a current representation for the target resource or is not willing to disclose that one exists.\"",
    },
    "405" : {
        "text": "Method Not Allowed",
        "description": "\"indicates that the method specified in the request-line is known by the origin server but not supported by the target resource.\"",
    },
    "406" : {
        "text": "Not Acceptable",
        "description": "\"indicates that the target resource does not have a current representation that would be acceptable to the user agent, according to the proactive negotiation header fields received in the request, and the server is unwilling to supply a default representation.\"",
    },
    "407" : {
        "text": "Proxy Authentication Required",
        "description": "\"is similar to 401 (Unauthorized), but indicates that the client needs to authenticate itself in order to use a proxy.\"",
    },
    "408" : {
        "text": "Request Timeout",
        "description": "\"indicates that the server did not receive a complete request message within the time that it was prepared to wait.\"",
    },
    "409" : {
        "text": "Conflict",
        "description": "\"indicates that the request could not be completed due to a conflict with the current state of the resource.\"",
    },
    "410" : {
        "text": "Gone",
        "description": "\"indicates that access to the target resource is no longer available at the origin server and that this condition is likely to be permanent.\"",
    },
    "411" : {
        "text": "Length Required",
        "description": "\"indicates that the server refuses to accept the request without a defined Content-Length.\"",
    },
    "412" : {
        "text": "Precondition Failed",
        "description": "\"indicates that one or more preconditions given in the request header fields evaluated to false when tested on the server.\"",
    },
    "413" : {
        "text": "Payload Too Large",
        "description": "\"indicates that the server is refusing to process a request because the request payload is larger than the server is willing or able to process.\"",
    },
    "414" : {
        "text": "URI Too Long",
        "description": "\"indicates that the server is refusing to service the request because the request-target is longer than the server is willing to interpret.\"",
    },
    "415" : {
        "text": "Unsupported Media Type",
        "description": "\"indicates that the origin server is refusing to service the request because the payload is in a format not supported by the target resource for this method.\"",
    },
    "416" : {
        "text": "Range Not Satisfiable",
        "description": "\"indicates that none of the ranges in the request's range header field overlap the current extent of the selected resource or that the set of ranges requested has been rejected due to invalid ranges or an excessive request of small or overlapping ranges.\"",
    },
    "417" : {
        "text": "Expectation Failed",
        "description": "\"indicates that the expectation given in the request's Expect header field could not be met by at least one of the inbound servers.\"",
    },
    "418" : {
        "text": "I'm a teapot",
        "description": "\"Any attempt to brew coffee with a teapot should result in the error code 418 I'm a teapot.\"",
    },
    "426" : {
        "text": "Upgrade Required",
        "description": "\"indicates that the server refuses to perform the request using the current protocol but might be willing to do so after the client upgrades to a different protocol.\"",
    },
    "422" : {
        "text": "Unprocessable Entity",
        "description": "\"means the server understands the content type of the request entity (hence a 415(Unsupported Media Type) status code is inappropriate), and the syntax of the request entity is correct (thus a 400 (Bad Request) status code is inappropriate) but was unable to process the contained instructions.\"",
    },
    "423" : {
        "text": "Locked",
        "description": "\"means the source or destination resource of a method is locked.\"",
    },
    "424" : {
        "text": "Failed Dependency",
        "description": "\"means that the method could not be performed on the resource because the requested action depended on another action and that action failed.\"",
    },
    "428" : {
        "text": "Precondition Required",
        "description": "\"indicates that the origin server requires the request to be conditional.\"",
    },
    "429" : {
        "text": "Too Many Requests",
        "description": "\"indicates that the user has sent too many requests in a given amount of time (\"rate limiting\").\"",
    },
    "431" : {
        "text": "Request Header Fields Too Large",
        "description": "\"indicates that the server is unwilling to process the request because its header fields are too large.\"",
    },
    "451" : {
        "text": "Unavailable For Legal Reasons",
        "description": "\"This status code indicates that the server is denying access to the resource in response to a legal demand.\"",
    },
    "500" : {
        "text": "Internal Server Error",
        "description": "\"indicates that the server encountered an unexpected condition that prevented it from fulfilling the request.\"",
    },
    "501" : {
        "text": "Not Implemented",
        "description": "\"indicates that the server does not support the functionality required to fulfill the request.\"",
    },
    "502" : {
        "text": "Bad Gateway",
        "description": "\"indicates that the server, while acting as a gateway or proxy, received an invalid response from an inbound server it accessed while attempting to fulfill the request.\"",
    },
    "503" : {
        "text": "Service Unavailable",
        "description": "\"indicates that the server is currently unable to handle the request due to a temporary overload or scheduled maintenance, which will likely be alleviated after some delay.\"",
    },
    "504" : {
        "text": "Gateway Time-out",
        "description": "\"indicates that the server, while acting as a gateway or proxy, did not receive a timely response from an upstream server it needed to access in order to complete the request.\"",
    },
    "505" : {
        "text": "HTTP Version Not Supported",
        "description": "\"indicates that the server does not support, or refuses to support, the protocol version that was used in the request message.\"",
    },
    "506" : {
        "text": "Variant Also Negotiates",
        "description": "\"indicates that the server has an internal configuration error: the chosen variant resource is configured to engage in transparent content negotiation itself, and is therefore not a proper end point in the negotiation process.\"",
    },
    "507" : {
        "text": "Insufficient Storage",
        "description": "\"means the method could not be performed on the resource because the server is unable to store the representation needed to successfully complete the request.\"",
    },
    "511" : {
        "text": "Network Authentication Required",
        "description": "\"indicates that the client needs to authenticate to gain network access.\"",
    }
}

'''
DEBUG - debug message
INFO - info message
WARNING - warn message
ERROR - error message
CRITICAL - critical message
'''
LOG_LEVEL_VALUE = {'DEBUG':10,'INFO':20,'WARNING':30,'ERROR':40,'CRITICAL':50}
if str(os.environ.get('PYMETRICK_LOG_LEVEL',None)).upper() in LOG_LEVEL_VALUE:
    LOG_LEVEL = LOG_LEVEL_VALUE[str(os.environ.get('PYMETRICK_LOG_LEVEL')).upper()]
else:
    LOG_LEVEL = 30
LOG_FILENAME = '-'.join([os.path.abspath(__file__).split(os.sep)[len(os.path.abspath(__file__).split(os.sep))-1],])[:-3]
LOG = logging.getLogger(LOG_FILENAME)

if 'LD_LIBRARY_PATH' in list(os.environ.keys()):
    # CGI environment
    sys.stdout = sys.stderr
    logging.basicConfig(stream = sys.stderr, level=LOG_LEVEL, format='%(filename)s[line:%(lineno)d] - %(levelname)s - %(message)s')
else:
    # not CGI environment
    logging.basicConfig(stream=sys.stderr)
    hdlr = logging.handlers.RotatingFileHandler(filename=LOG_FILENAME+'.log',mode='a', encoding='utf-8', maxBytes=1048576, backupCount=3)
    formatter = logging.Formatter('%(asctime)s - %(filename)s[line:%(lineno)d] - %(levelname)s - %(message)s')
    hdlr.setFormatter(formatter)
    LOG.addHandler(hdlr)
    LOG.setLevel(LOG_LEVEL)

bytes_types = (bytes, bytearray)

class wsgateway:
    '''Gestion de entorno Web Server Gateway Interface'''

    def __init__(self, environment, start_response, settings = None):

        try:
            self._p = None
            self._q = None
            self.status = '200'
            self.header = dict()
            self.parametros = None
            self.environ = environment
            self.settings = settings if settings is not None else {};
            self.response = start_response
            if os.environ.get('PYMETRICK_LOG',None) == 'logging.DEBUG':
                '''mostrar entorno en caso de DEBUG
                   con claves ordenadas
                '''
                keys = sorted(list(self.environ.keys()))
                for key in keys:
                    LOG.debug("%s: %s" % (key, self.environ[key]))

            # check xmlhttprequest
            self.ajax_request = self.is_xmlhttprequest()
            # check REQUEST_METHOD
            if self.is_allowed_request_method():
                '''La variable de environ CONTENT_LENGTH puede ser 0 o estar ausente en las variables de environ
                   esta relacionada con el contenido de wsgi.input
                '''
                if int(self.environ.get('CONTENT_LENGTH','0'))>0 and self.environ.get('wsgi.input',None):
                    self._p = self.environ['wsgi.input'].read(int(self.environ['CONTENT_LENGTH']))
                else:
                    '''QUERY_STRING puede no tener contenido o estar ausente en las variables de environ'''
                    try:
                        if self.environ.get('QUERY_STRING',None):
                            '''QUERY_STRING The portion of the request URL that follows the "?" , if any. May be empty or absent.'''
                            self._p = self.environ.get('QUERY_STRING','')
                    except Exception as e:
                        pass

                self.parametros = None
                if self._p is not None:
                    self.request()
                else:
                    if os.environ.get('PYMETRICK_LOG',None) == 'logging.DEBUG':
                        if self._p is not None:
                            LOG.debug("self.parametros %s" % (self._p,))
                    self.parametros = None

        except Exception as e:
            tb = sys.exc_info()[2]
            LOG.error("Error en metodo '__init__' <%s> en linea %s !!!\n" % (repr(e),tb.tb_lineno))

    def is_xmlhttprequest(self):
        '''Check xmlhttprequest'''
        if self.environ.get('HTTP_X_REQUESTED_WITH','').lower() == 'xmlhttprequest' or self.environ.get('X_REQUESTED_WITH','').lower() == 'xmlhttprequest':
            return True
        return False

    def is_allowed_request_method(self):
        '''Check REQUEST_METHOD'''
        if self.environ.get('REQUEST_METHOD','').upper() in ['GET','POST','PUT','DELETE','HEADER']:
            return True
        return False

    def get_key(self, key):
        return self.environment.get(key)

    def get_request(self):
        return self.parametros

    def set_status(self,status):
        self.status = status

    def set_header(self,header):
        if isinstance(header,(list,tuple)):
            for n in header:
                if isinstance(n,(list,tuple)):
                    if len(n)==2:
                        self.header[n[0]] = n[1]
                elif len(header)==2:
                    self.header[header[0]] = header[1]
                    break

    def status_response(self,output,jwt):
        try:
            self.header_list = []
            LOG.debug('self.status {0}'.format(self.status))
            LOG.debug('self.status type {0}'.format(type(self.status).__name__))
            if self.header.get('content-type',None):
                LOG.debug('header[content-type] {0}'.format(self.header['content-type']))
                LOG.debug('header[content-type] type {0}'.format(type(self.header['content-type']).__name__))
            if str(self.status) in STATUS_CODE:
                self.status = getStatus(self.status)
                self.header['X-Authorization'] = jwt.get_token()
                self.header['Set-Cookie'] = jwt.get_cookie()
                if self.header.get('content-type',None):
                    if self.header['content-type'] == 'application/octet-stream' or self.header['content-type'] == 'application/force-download':
                        # corregir content-type
                        if output[output.find('.')+1:].replace('_','') in CONTENT_TYPE.keys():
                            self.header['content-type'] = CONTENT_TYPE[output[output.find('.')+1:].replace('_','')]
                        self.header['content-transfer-encoding'] = 'binary'
                        self.header['content-disposition'] =  'attachment; filename={0};'.format(output)
                        file_path = os.path.join(self.settings.get('download',''), bytes2str(output))
                        self.header['X-Sendfile'] = file_path
                        for key,value in self.header.items():
                            self.header_list.append((key,value,))
                        self.response(self.status,self.header_list)
                        return [b'']
                else:
                    self.header['content-type'] = 'text/html; charset=UTF-8'
                self.header['content-length'] = str(len(str2bytes(output)))
                self.header['Cache-Control'] = 'no-store'
                for key,value in self.header.items():
                    self.header_list.append((key,value,))
                self.response(self.status,self.header_list)
                if self.environ.get('REQUEST_METHOD','').upper() == 'HEADER':
                    return [b'']
                else:
                    return str2bytes([output])
        except Exception as e:
            tb = sys.exc_info()[2]
            LOG.error("Error en metodo 'status_response' <%s> en linea %s !!!\n" % (repr(e),tb.tb_lineno))


    def request(self):
        ''' Process all parameters except content-type <multipart/form-data>'''
        try:
            # self.environ.get('CONTENT_TYPE','').split(';')[0] in ('multipart/form-data',) &&
            if self.environ['REQUEST_METHOD'] == 'POST' and self.environ.get('CONTENT_TYPE','').split(';')[0] in ('multipart/form-data',):
                # request for file upload
                # get first 'boundary' code
                _q_list = self._p.split(b"\r",1)
                json_q = dict()
                # split by 'boundary'
                _q_list = self._p.split(_q_list[0])

                _pp = dict()
                for n in range(len(_q_list)):
                    json_q[n] = _q_list[n].split(b"\r\n")
                    if len(json_q[n])>3:
                        # json_q[n][1]  Content-Disposition: form-data; name="file"; filename="example.jpg"
                        # json_q[n])>2  Content-Type: image/jpeg if a file or None
                        # json_q[n][3]  Value
                        if json_q[n][2].lower().startswith(b'content-type'):
                            content_disposition_name = None             # id field
                            content_disposition_filename = None         # file name
                            # join list like file
                            json_q[n][4] += b"\r\n"
                            for m in range(5,len(json_q[n])):
                                json_q[n][4] += b''.join([json_q[n][m],b"\r\n"])
                            # id field
                            content_disposition_name = json_q[n][1].split(b':')[1].split(b';')[1].split(b'=')[1].replace(b'"', b'')
                            # file name
                            content_disposition_filename = json_q[n][1].split(b':')[1].split(b';')[2].split(b'=')[1].replace(b'"', b'')
                            if content_disposition_name is not None and content_disposition_filename is not None:
                                # if form is multi files, file name save like list
                                if _pp.get(content_disposition_name,None):
                                    _pp[content_disposition_name].append(content_disposition_filename)
                                else:
                                    _pp[content_disposition_name]= [content_disposition_filename]
                            # save file on self.settings['upload']
                            if len(json_q[n][4])>0:
                                file_path = os.path.join(self.settings.get('upload',''), bytes2str(content_disposition_filename))
                                counter = 0
                                with open(file_path, 'wb') as output_file:
                                    output_file.write(json_q[n][4])
                                output_file.closed
                        else:
                            # is a value
                            content_disposition_name = None
                            content_type = None
                            for m in range(len(json_q[n])):
                                if json_q[n][m].lower().startswith(b'content-disposition'):
                                    # id field
                                    content_disposition_name = json_q[n][1].split(b':')[1].split(b';')[1].split(b'=')[1].replace(b'"', b'')
                                elif len(json_q[n][m])>0:
                                    # value
                                    content_type = json_q[n][3].strip()
                            if content_disposition_name is not None and content_type is not None:
                                _pp[content_disposition_name]=content_type
                # set clear memory
                _q_list = None
                json_q = None
            else:
                # request for json or scheme key=value&
                try:
                    LOG.debug('self._p %s' % self._p)
                    for n in self.environ.keys():
                        LOG.debug("key %s  value %s" % (n,self.environ[n]))
                    if isinstance(self._p, (str,)) and (re.search('^\{.*}$', self._p ) or re.search('^\[{.*}]$', self._p )):
                        # check if parameters like str received with format JSON (dict, array)
                        LOG.debug("json str")
                        json_q = json.loads(self._p)
                        LOG.debug("json str json_q %s" % (json.dumps(json_q),))
                    elif isinstance(self._p, (bytes,bytearray)) and (re.search(b'^\{.*}$', self._p ) or re.search(b'^\[{.*}]$', self._p )):
                        # check if parameters like bytes received with format JSON (dict, array)
                        LOG.debug("json bytes")
                        json_q = json.loads(self._p)
                        LOG.debug("json bytes json_q %s" % (json.dumps(json_q),))
                    else:
                        # parameters received like scheme key=value
                        LOG.debug("other")
                        json_q = parse_qs(self._p)
                        LOG.debug("other json_q %s" % (json.dumps(json_q),))
                except Exception as e:
                    LOG.error("parameters error %s" % (self._p,))
                    json_q = parse_qs(self._p)
                # check if parameters value coded on base64 y decode them
                json_q = obj_b64decode(json_q)
                # Change from bytes to str
                json_q = bytes2str(json_q)

                self.parametros = json.dumps(json_q, default=to_json)
        except Exception as e:
            tb = sys.exc_info()[2]
            LOG.error("metodo 'request' <%s> en linea %s !!!\n" % (repr(e),tb.tb_lineno))

    def request_multipart(self):
        ''' Process parameters content-type <multipart/form-data>'''
        try:
            if self.environ['REQUEST_METHOD'] == 'POST':
                # get first 'boundary' code
                _q_list = self._p.split(b"\r",1)
                json_q = dict()
                # split by 'boundary'
                _q_list = self._p.split(_q_list[0])

                _pp = dict()
                for n in range(len(_q_list)):
                    json_q[n] = _q_list[n].split(b"\r\n")
                    if len(json_q[n])>3:
                        # json_q[n][1]  Content-Disposition: form-data; name="file"; filename="example.jpg"
                        # json_q[n])>2  Content-Type: image/jpeg if a file or None
                        # json_q[n][3]  Value
                        if json_q[n][2].lower().startswith(b'content-type'):
                            content_disposition_name = None             # id field
                            content_disposition_filename = None         # file name
                            # join list like file
                            json_q[n][4] += b"\r\n"
                            for m in range(5,len(json_q[n])):
                                json_q[n][4] += b''.join([json_q[n][m],b"\r\n"])
                            # id field
                            content_disposition_name = json_q[n][1].split(b':')[1].split(b';')[1].split(b'=')[1].replace(b'"', b'')
                            # file name
                            content_disposition_filename = json_q[n][1].split(b':')[1].split(b';')[2].split(b'=')[1].replace(b'"', b'')
                            if content_disposition_name is not None and content_disposition_filename is not None:
                                # if form is multi files, file name save like list
                                if _pp.get(content_disposition_name,None):
                                    _pp[content_disposition_name].append(content_disposition_filename)
                                else:
                                    _pp[content_disposition_name]= [content_disposition_filename]
                            # save file on self.settings['upload']
                            if len(json_q[n][4])>0:
                                file_path = os.path.join(self.settings.get('upload',''), bytes2str(content_disposition_filename))
                                counter = 0
                                with open(file_path, 'wb') as output_file:
                                    output_file.write(json_q[n][4])
                                output_file.closed
                        else:
                            # is a value
                            content_disposition_name = None
                            content_type = None
                            for m in range(len(json_q[n])):
                                if json_q[n][m].lower().startswith(b'content-disposition'):
                                    # id field
                                    content_disposition_name = json_q[n][1].split(b':')[1].split(b';')[1].split(b'=')[1].replace(b'"', b'')
                                elif len(json_q[n][m])>0:
                                    # value
                                    content_type = json_q[n][3].strip()
                            if content_disposition_name is not None and content_type is not None:
                                _pp[content_disposition_name]=content_type
                # set clear memory
                _q_list = None
                json_q = None
        except Exception as e:
            tb = sys.exc_info()[2]
            LOG.error("metodo 'request_multipart' <%s> en linea %s !!!\n" % (repr(e),tb.tb_lineno))

def getStatus(status_id):
    '''Process LOT errors'''
    try:
        if isinstance(status_id,(int,)):
            status_id = str(status_id)
        if STATUS_CODE.get(status_id,None):
            return "{0} {1}".format(status_id,STATUS_CODE[status_id]['text'])
        else:
            status_id = "200"
            return "{0} {1}".format(status_id,STATUS_CODE[status_id]['text'])
    except Exception as e:
        tb = sys.exc_info()[2]
        LOG.error("Error <%s> en linea %s !!!\n" % (repr(e),tb.tb_lineno))

_implicit_encoding = 'ascii'
_implicit_errors = 'strict'

def _noop(obj):
    return obj

def _encode_result(obj, encoding=_implicit_encoding,
                        errors=_implicit_errors):
    return obj.encode(encoding, errors)

def _decode_args(args, encoding=_implicit_encoding,
                       errors=_implicit_errors):
    return tuple(x.decode(encoding, errors) if x else '' for x in args)

def _coerce_args(*args):
    # Invokes decode if necessary to create str args
    # and returns the coerced inputs along with
    # an appropriate result coercion function
    #   - noop for str inputs
    #   - encoding function otherwise
    str_input = isinstance(args[0], str)
    for arg in args[1:]:
        # We special-case the empty string to support the
        # "scheme=''" default argument to some functions
        if arg and isinstance(arg, str) != str_input:
            raise TypeError("Cannot mix str and non-str arguments")
    if str_input:
        return args + (_noop,)
    return _decode_args(args) + (_encode_result,)

_hexdig = '0123456789ABCDEFabcdef'
_hextobyte = None

def unquote_to_bytes(string):
    """unquote_to_bytes('abc%20def') -> b'abc def'."""
    # Note: strings are encoded as UTF-8. This is only an issue if it contains
    # unescaped non-ASCII characters, which URIs should not.
    if not string:
        # Is it a string-like object?
        string.split
        return b''
    if isinstance(string, str):
        string = string.encode('utf-8')
    bits = string.split(b'%')
    if len(bits) == 1:
        return string
    res = [bits[0]]
    append = res.append
    # Delay the initialization of the table to not waste memory
    # if the function is never called
    global _hextobyte
    if _hextobyte is None:
        _hextobyte = {(a + b).encode(): bytes.fromhex(a + b)
                      for a in _hexdig for b in _hexdig}
    for item in bits[1:]:
        try:
            append(_hextobyte[item[:2]])
            append(item[2:])
        except KeyError:
            append(b'%')
            append(item)
    return b''.join(res)

_asciire = re.compile('([\x00-\x7f]+)')

def unquote(string, encoding='utf-8', errors='replace'):
    """Replace %xx escapes by their single-character equivalent. The optional
    encoding and errors parameters specify how to decode percent-encoded
    sequences into Unicode characters, as accepted by the bytes.decode()
    method.
    By default, percent-encoded sequences are decoded with UTF-8, and invalid
    sequences are replaced by a placeholder character.
    unquote('abc%20def') -> 'abc def'.
    """
    if '%' not in string:
        string.split
        return string
    if encoding is None:
        encoding = 'utf-8'
    if errors is None:
        errors = 'replace'
    bits = _asciire.split(string)
    res = [bits[0]]
    append = res.append
    for i in range(1, len(bits), 2):
        append(unquote_to_bytes(bits[i]).decode(encoding, errors))
        append(bits[i + 1])
    return ''.join(res)

def parse_qs(qs, keep_blank_values=False, strict_parsing=False,encoding='utf-8', errors='replace'):
    """Parse a query given as a string argument.
        Arguments:
        qs: percent-encoded query string to be parsed
        keep_blank_values: flag indicating whether blank values in
            percent-encoded queries should be treated as blank strings.
            A true value indicates that blanks should be retained as
            blank strings.  The default false value indicates that
            blank values are to be ignored and treated as if they were
            not included.
        strict_parsing: flag indicating what to do with parsing errors.
            If false (the default), errors are silently ignored.
            If true, errors raise a ValueError exception.
        encoding and errors: specify how to decode percent-encoded sequences
            into Unicode characters, as accepted by the bytes.decode() method.
        Returns a dictionary of list - Modify to return a dictionary of values
    """
    parsed_result = {}
    pairs = parse_qsl(qs, keep_blank_values, strict_parsing, encoding=encoding, errors=errors)
    for name, value in pairs:
        if name in parsed_result:
            # if more than 1 value
            if isinstance(parsed_result[name],(str,bytes,bytearray)):
                parsed_tmp = parsed_result[name]
                parsed_result[name] = [parsed_tmp]
            if isinstance(parsed_result[name],(list,tuple)):
                parsed_result[name].append(value)
        else:
            # if only 1 value
            parsed_result[name] = value
    return parsed_result

def parse_qsl(qs, keep_blank_values=False, strict_parsing=False, encoding='utf-8', errors='replace'):
    """Parse a query given as a string argument.
        Arguments:
        qs: percent-encoded query string to be parsed
        keep_blank_values: flag indicating whether blank values in
            percent-encoded queries should be treated as blank strings.
            A true value indicates that blanks should be retained as blank
            strings.  The default false value indicates that blank values
            are to be ignored and treated as if they were  not included.
        strict_parsing: flag indicating what to do with parsing errors. If
            false (the default), errors are silently ignored. If true,
            errors raise a ValueError exception.
        encoding and errors: specify how to decode percent-encoded sequences
            into Unicode characters, as accepted by the bytes.decode() method.
        Returns a list, as G-d intended.
    """
    qs, _coerce_result = _coerce_args(qs)
    pairs = [s2 for s1 in qs.split('&') for s2 in s1.split(';')]
    r = []
    for name_value in pairs:
        if not name_value and not strict_parsing:
            continue
        nv = name_value.split('=', 1)
        if len(nv) != 2:
            if strict_parsing:
                raise ValueError("bad query field: %r" % (name_value,))
            # Handle case of a control-name with no equal sign
            if keep_blank_values:
                nv.append('')
            else:
                continue
        if len(nv[1]) or keep_blank_values:
            name = nv[0].replace('+', ' ')
            name = unquote(name, encoding=encoding, errors=errors)
            name = _coerce_result(name)
            value = nv[1].replace('+', ' ')
            value = unquote(value, encoding=encoding, errors=errors)
            value = _coerce_result(value)
            r.append((name, value))
    return r

def unquote_plus(string, encoding='utf-8', errors='replace'):
    """Like unquote(), but also replace plus signs by spaces, as required for
    unquoting HTML form values.
    unquote_plus('%7e/abc+def') -> '~/abc def'
    """
    string = string.replace('+', ' ')
    return unquote(string, encoding, errors)

_ALWAYS_SAFE = frozenset(b'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
                         b'abcdefghijklmnopqrstuvwxyz'
                         b'0123456789'
                         b'_.-~')
_ALWAYS_SAFE_BYTES = bytes(_ALWAYS_SAFE)
_safe_quoters = {}

class Quoter(collections.defaultdict):
    """A mapping from bytes (in range(0,256)) to strings.
    String values are percent-encoded byte values, unless the key < 128, and
    in the "safe" set (either the specified safe set, or default set).
    """
    # Keeps a cache internally, using defaultdict, for efficiency (lookups
    # of cached keys don't call Python code at all).
    def __init__(self, safe):
        """safe: bytes object."""
        self.safe = _ALWAYS_SAFE.union(safe)

    def __repr__(self):
        # Without this, will just display as a defaultdict
        return "<%s %r>" % (self.__class__.__name__, dict(self))

    def __missing__(self, b):
        # Handle a cache miss. Store quoted string in cache and return.
        res = chr(b) if b in self.safe else '%{:02X}'.format(b)
        self[b] = res
        return res

def quote(string, safe='/', encoding=None, errors=None):
    """quote('abc def') -> 'abc%20def'
    Each part of a URL, e.g. the path info, the query, etc., has a
    different set of reserved characters that must be quoted.
    RFC 3986 Uniform Resource Identifiers (URI): Generic Syntax lists
    the following reserved characters.
    reserved    = ";" | "/" | "?" | ":" | "@" | "&" | "=" | "+" |
                  "$" | "," | "~"
    Each of these characters is reserved in some component of a URL,
    but not necessarily in all of them.
    Python 3.7 updates from using RFC 2396 to RFC 3986 to quote URL strings.
    Now, "~" is included in the set of reserved characters.
    By default, the quote function is intended for quoting the path
    section of a URL.  Thus, it will not encode '/'.  This character
    is reserved, but in typical usage the quote function is being
    called on a path where the existing slash characters are used as
    reserved characters.
    string and safe may be either str or bytes objects. encoding and errors
    must not be specified if string is a bytes object.
    The optional encoding and errors parameters specify how to deal with
    non-ASCII characters, as accepted by the str.encode method.
    By default, encoding='utf-8' (characters are encoded with UTF-8), and
    errors='strict' (unsupported characters raise a UnicodeEncodeError).
    """
    if isinstance(string, str):
        if not string:
            return string
        if encoding is None:
            encoding = 'utf-8'
        if errors is None:
            errors = 'strict'
        string = string.encode(encoding, errors)
    else:
        if encoding is not None:
            raise TypeError("quote() doesn't support 'encoding' for bytes")
        if errors is not None:
            raise TypeError("quote() doesn't support 'errors' for bytes")
    return quote_from_bytes(string, safe)

def quote_plus(string, safe='', encoding=None, errors=None):
    """Like quote(), but also replace ' ' with '+', as required for quoting
    HTML form values. Plus signs in the original string are escaped unless
    they are included in safe. It also does not have safe default to '/'.
    """
    # Check if ' ' in string, where string may either be a str or bytes.  If
    # there are no spaces, the regular quote will produce the right answer.
    if ((isinstance(string, str) and ' ' not in string) or
        (isinstance(string, bytes) and b' ' not in string)):
        return quote(string, safe, encoding, errors)
    if isinstance(safe, str):
        space = ' '
    else:
        space = b' '
    string = quote(string, safe + space, encoding, errors)
    return string.replace(' ', '+')

def quote_from_bytes(bs, safe='/'):
    """Like quote(), but accepts a bytes object rather than a str, and does
    not perform string-to-bytes encoding.  It always returns an ASCII string.
    quote_from_bytes(b'abc def\x3f') -> 'abc%20def%3f'
    """
    if not isinstance(bs, bytes_types):
        raise TypeError("quote_from_bytes() expected bytes")
    if not bs:
        return ''
    if isinstance(safe, str):
        # Normalize 'safe' by converting to bytes and removing non-ASCII chars
        safe = safe.encode('ascii', 'ignore')
    else:
        safe = bytes([c for c in safe if c < 128])
    if not bs.rstrip(_ALWAYS_SAFE_BYTES + safe):
        return bs.decode()
    try:
        quoter = _safe_quoters[safe]
    except KeyError:
        _safe_quoters[safe] = quoter = Quoter(safe).__getitem__
    return ''.join([quoter(char) for char in bs])

if __name__  == "__main__":
    print ('''copyright {0}'''.format( __copyright__))
    print ('''license {0}'''.format( __license__))
    print ('''version {0}'''.format( __version__))
    if len(sys.argv) < 2:
        sys.stderr.write("for help use -h o --help")
    elif sys.argv[1]=='-h' or sys.argv[1]=='--help':
        print ('''
        Tratamiento de WSGI\n\n''')










#!/usr/bin/env python
"""
geoipy.py
A simple IP geolocation python script.
Uses geody.com geolocation web service.
Requires BeautifulSoup library.

Example:
$ ./geoipy.py 198.117.0.122
IP: 198.117.0.122 Location: MSFC, ALABAMA, \
United States (National Aeronautics and Space Administration)

ksaver, June 18, 2011.
Public Domain Code.
"""
import sys
# sin bytecode
sys.dont_write_bytecode = True
import re
from urllib import urlopen
from BeautifulSoup import BeautifulSoup as Soup

def main(ipaddr):
    """Geo-locates an IP address passed in argv[1]."""

    geody = "http://www.geody.com/geoip.php?ip=" + ipaddr
    html_page = urlopen(geody).read()
    soup = Soup(html_page)

    # Filter paragraph containing geolocation info.
    paragraph = soup('p')[3]

    # Remove html tags using regex.
    geo_txt = re.sub(r'<.*?>', '', str(paragraph))
    print (geo_txt[32:].strip() + '\n')

if __name__ == "__main__":
    if len(sys.argv) > 1:
        main(sys.argv[1])
    else:
        main('127.0.0.1')



https://github.com/danielmiessler/SecLists/blob/master/Passwords/

twttr.BANNED_PASSWORDS = [
    "000000",
    "111111",
    "11111111",
    "112233",
    "121212",
    "123123",
    "123456",
    "1234567",
    "12345678",
    "123456789",
    "131313",
    "232323",
    "654321",
    "666666",
    "696969",
    "777777",
    "7777777",
    "8675309",
    "987654",
    "aaaaaa",
    "abc123",
    "abc123",
    "abcdef",
    "abgrtyu",
    "access",
    "access14",
    "action",
    "albert",
    "alberto",
    "alexis",
    "alejandra",
    "alejandro",
    "amanda",
    "amateur",
    "america",
    "andrea",
    "andrew",
    "angela",
    "angels",
    "animal",
    "anthony",
    "apollo",
    "apples",
    "arsenal",
    "arthur",
    "asdfgh",
    "asdfgh",
    "ashley",
    "asshole",
    "august",
    "austin",
    "badboy",
    "bailey",
    "banana",
    "barney",
    "baseball",
    "batman",
    "beatriz",
    "beaver",
    "beavis",
    "bigcock",
    "bigdaddy",
    "bigdick",
    "bigdog",
    "bigtits",
    "birdie",
    "bitches",
    "biteme",
    "blazer",
    "blonde",
    "blondes",
    "blowjob",
    "blowme",
    "bond007",
    "bonita",
    "bonnie",
    "booboo",
    "booger",
    "boomer",
    "boston",
    "brandon",
    "brandy",
    "braves",
    "brazil",
    "bronco",
    "broncos",
    "bulldog",
    "buster",
    "butter",
    "butthead",
    "calvin",
    "camaro",
    "cameron",
    "canada",
    "captain",
    "carlos",
    "carter",
    "casper",
    "charles",
    "charlie",
    "cheese",
    "chelsea",
    "chester",
    "chicago",
    "chicken",
    "cocacola",
    "coffee",
    "college",
    "compaq",
    "computer",
    "consumer",
    "cookie",
    "cooper",
    "corvette",
    "cowboy",
    "cowboys",
    "crystal",
    "cumming",
    "cumshot",
    "dakota",
    "dallas",
    "daniel",
    "danielle",
    "debbie",
    "dennis",
    "diablo",
    "diamond",
    "doctor",
    "doggie",
    "dolphin",
    "dolphins",
    "donald",
    "dragon",
    "dreams",
    "driver",
    "eagle1",
    "eagles",
    "edward",
    "einstein",
    "erotic",
    "estrella",
    "extreme",
    "falcon",
    "fender",
    "ferrari",
    "firebird",
    "fishing",
    "florida",
    "flower",
    "flyers",
    "football",
    "forever",
    "freddy",
    "freedom",
    "fucked",
    "fucker",
    "fucking",
    "fuckme",
    "fuckyou",
    "gandalf",
    "gateway",
    "gators",
    "gemini",
    "george",
    "giants",
    "ginger",
    "gizmodo",
    "golden",
    "golfer",
    "gordon",
    "gregory",
    "guitar",
    "gunner",
    "hammer",
    "hannah",
    "hardcore",
    "harley",
    "heather",
    "helpme",
    "hentai",
    "hockey",
    "hooters",
    "horney",
    "hotdog",
    "hunter",
    "hunting",
    "iceman",
    "iloveyou",
    "internet",
    "iwantu",
    "jackie",
    "jackson",
    "jaguar",
    "jasmine",
    "jasper",
    "jennifer",
    "jeremy",
    "jessica",
    "johnny",
    "johnson",
    "jordan",
    "joseph",
    "joshua",
    "junior",
    "justin",
    "killer",
    "knight",
    "ladies",
    "lakers",
    "lauren",
    "leather",
    "legend",
    "letmein",
    "letmein",
    "little",
    "london",
    "lovers",
    "maddog",
    "madison",
    "maggie",
    "magnum",
    "marine",
    "mariposa",
    "marlboro",
    "martin",
    "marvin",
    "master",
    "matrix",
    "matthew",
    "maverick",
    "maxwell",
    "melissa",
    "member",
    "mercedes",
    "merlin",
    "michael",
    "michelle",
    "mickey",
    "midnight",
    "miller",
    "mistress",
    "monica",
    "monkey",
    "monkey",
    "monster",
    "morgan",
    "mother",
    "mountain",
    "muffin",
    "murphy",
    "mustang",
    "naked",
    "nascar",
    "nathan",
    "naughty",
    "ncc1701",
    "newyork",
    "nicholas",
    "nicole",
    "nipple",
    "nipples",
    "oliver",
    "orange",
    "packers",
    "panther",
    "panties",
    "parker",
    "password",
    "password",
    "password1",
    "password12",
    "password123",
    "patrick",
    "peaches",
    "peanut",
    "pepper",
    "phantom",
    "phoenix",
    "player",
    "please",
    "pookie",
    "porsche",
    "prince",
    "princess",
    "private",
    "purple",
    "pussies",
    "qazwsx",
    "qwerty",
    "qwertyui",
    "rabbit",
    "rachel",
    "racing",
    "raiders",
    "rainbow",
    "ranger",
    "rangers",
    "rebecca",
    "redskins",
    "redsox",
    "redwings",
    "richard",
    "robert",
    "roberto",
    "rocket",
    "rosebud",
    "runner",
    "rush2112",
    "russia",
    "samantha",
    "sammy",
    "samson",
    "sandra",
    "saturn",
    "scooby",
    "scooter",
    "scorpio",
    "scorpion",
    "sebastian",
    "secret",
    "sexsex",
    "shadow",
    "shannon",
    "shaved",
    "sierra",
    "silver",
    "skippy",
    "slayer",
    "smokey",
    "snoopy",
    "soccer",
    "sophie",
    "spanky",
    "sparky",
    "spider",
    "squirt",
    "srinivas",
    "startrek",
    "starwars",
    "steelers",
    "steven",
    "sticky",
    "stupid",
    "success",
    "suckit",
    "summer",
    "sunshine",
    "superman",
    "surfer",
    "swimming",
    "sydney",
    "tequiero",
    "taylor",
    "tennis",
    "teresa",
    "tester",
    "testing",
    "theman",
    "thomas",
    "thunder",
    "thx1138",
    "tiffany",
    "tigers",
    "tigger",
    "tomcat",
    "topgun",
    "toyota",
    "travis",
    "trouble",
    "trustno1",
    "tucker",
    "turtle",
    "twitter",
    "united",
    "vagina",
    "victor",
    "victoria",
    "viking",
    "voodoo",
    "voyager",
    "walter",
    "warrior",
    "welcome",
    "whatever",
    "william",
    "willie",
    "wilson",
    "winner",
    "winston",
    "winter",
    "wizard",
    "xavier",
    "xxxxxx",
    "xxxxxxxx",
    "yamaha",
    "yankee",
    "yankees",
    "yellow",
    "zxcvbn",
    "zxcvbnm",
    "zzzzzz"
];